# -*- coding: utf-8 -*-
"""Structure with state curated by attribute descriptors."""

from abc import abstractmethod
from inspect import getmro
from typing import TYPE_CHECKING, TypeVar, overload
from weakref import WeakKeyDictionary

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

from six import iteritems, raise_from, string_types, with_metaclass

from .._bases import (
    ABSTRACT_TAG,
    FINAL_METHOD_TAG,
    MISSING,
    BaseDict,
    BaseHashable,
    BaseList,
    BaseMeta,
    BaseSet,
    Generic,
    final,
)
from .._states import DictState, SetState
from ..utils.custom_repr import custom_iterable_repr, custom_mapping_repr
from ..utils.factoring import format_factory, import_factory, run_factory
from ..utils.recursive_repr import recursive_repr
from ..utils.reraise_context import ReraiseContext
from ..utils.type_checking import assert_is_instance
from .bases import (
    BaseInteractiveStructure,
    BaseMutableStructure,
    BaseRelationship,
    BaseStructure,
    BaseStructureMeta,
)

if TYPE_CHECKING:
    from typing import (
        Any,
        Dict,
        Iterable,
        Iterator,
        Mapping,
        MutableMapping,
        Optional,
        Tuple,
        Type,
    )

    from ..utils.factoring import LazyFactory

__all__ = [
    "BaseAttributeMeta",
    "BaseAttribute",
    "BaseAttributeStructureMeta",
    "BaseAttributeStructure",
    "BaseInteractiveAttributeStructure",
    "BaseMutableAttributeStructure",
]


T = TypeVar("T")  # Any type.


class BaseAttributeMeta(BaseMeta):
    """Metaclass for `BaseAttribute`."""

    @property
    @abstractmethod
    def _relationship_type(cls):
        # type: () -> Type[BaseRelationship]
        """Relationship type."""
        return BaseRelationship


# noinspection PyTypeChecker
_BA = TypeVar("_BA", bound="BaseAttribute")


class BaseAttribute(with_metaclass(BaseAttributeMeta, BaseHashable, Generic[T])):
    """
    Base attribute descriptor for :class:`objetto.bases.BaseAttributeStructure` classes.

    Inherits from:
      - :class:`objetto.bases.BaseHashable`
      - :class:`typing.Generic`

    Inherited by:
      - :class:`objetto.data.DataAttribute`
      - :class:`objetto.objects.Attribute`

    :param relationship: Relationship.
    :type relationship: objetto.bases.BaseRelationship

    :param default: Default value.

    :param default_factory: Default value factory.
    :type default_factory: str or collections.abc.Callable or None

    :param module: Optional module path to use in case partial paths are provided.
    :type module: str or None

    :param required: Whether attribute is required to have a value or not.
    :type required: bool

    :param changeable: Whether attribute value can be changed.
    :type changeable: bool

    :param deletable: Whether attribute value can be deleted.
    :type deletable: bool

    :param finalized: If True, attribute can't be overridden by subclasses.
    :type finalized: bool

    :param abstracted: If True, attribute needs to be overridden by subclasses.
    :type abstracted: bool

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    :raises ValueError: Can't specify both 'default' and 'default_factory' arguments.
    :raises ValueError: Can't be 'required' and 'deletable' at the same time.
    :raises ValueError: Can't be 'finalized' and 'abstracted' at the same time.
    """

    __slots__ = (
        "__relationship",
        "__default",
        "__default_factory",
        "__module",
        "__required",
        "_changeable",
        "_deletable",
        "__finalized",
        "__abstracted",
        ABSTRACT_TAG,
        FINAL_METHOD_TAG,
    )

    def __init__(
        self,
        relationship=BaseRelationship(),  # type: BaseRelationship
        default=MISSING,  # type: Any
        default_factory=None,  # type: LazyFactory
        module=None,  # type: Optional[str]
        required=True,  # type: bool
        changeable=True,  # type: bool
        deletable=False,  # type: bool
        finalized=False,  # type: bool
        abstracted=False,  # type: bool
    ):
        # type: (...) -> None
        cls = type(self)

        # 'relationship'
        with ReraiseContext(TypeError, "'relationship' parameter"):
            assert_is_instance(relationship, cls._relationship_type, subtypes=False)

        # 'module'
        if module is None:
            module = relationship.module
        else:
            with ReraiseContext(TypeError, "'module' parameter"):
                assert_is_instance(module, string_types + (None,))
            module = module or None

        # 'default' and 'default_factory'
        if default is not MISSING and default_factory is not None:
            error = "can't specify both 'default' and 'default_factory' arguments"
            raise ValueError(error)
        with ReraiseContext((ValueError, TypeError), "'default_factory' parameter"):
            default_factory = format_factory(default_factory, module=module)

        # 'deletable' and 'required'
        if deletable and required:
            error = "can't be 'required' and 'deletable' at the same time"
            raise ValueError(error)

        # 'finalized' and 'abstracted'
        if finalized and abstracted:
            error = "can't be 'finalized' and 'abstracted' at the same time"
            raise ValueError(error)

        self.__relationship = relationship
        self.__default = default
        self.__default_factory = default_factory
        self.__module = module
        self.__required = bool(required)
        self._changeable = bool(changeable)
        self._deletable = bool(deletable)
        self.__finalized = bool(finalized)
        self.__abstracted = bool(abstracted)

        if finalized:
            setattr(self, FINAL_METHOD_TAG, True)
        if abstracted:
            setattr(self, ABSTRACT_TAG, True)

    @overload
    def __get__(self, instance, owner):
        # type: (_BA, None, Type[BaseAttributeStructure]) -> _BA
        pass

    @overload
    def __get__(self, instance, owner):
        # type: (BaseAttributeStructure, Type[BaseAttributeStructure]) -> T
        pass

    @overload
    def __get__(self, instance, owner):
        # type: (_BA, object, type) -> _BA
        pass

    def __get__(self, instance, owner):
        """
        Get attribute value when accessing from valid instance or this descriptor
        otherwise.

        :param instance: Instance.
        :type instance: objetto.bases.BaseAttributeStructure or None

        :param owner: Owner class.
        :type owner: type[objetto.bases.BaseAttributeStructure]

        :return: Value or this descriptor.
        :rtype: Any or objetto.bases.BaseAttribute
        """
        if instance is not None:
            attribute_type = getattr(type(instance), "_attribute_type", None)
            if attribute_type is not None and isinstance(self, attribute_type):
                return self.get_value(instance)
        return self

    @final
    def __hash__(self):
        # type: () -> int
        """
        Get hash based on object id.

        :return: Hash based on object id.
        :rtype: int
        """
        return hash(id(self))

    @final
    def __eq__(self, other):
        # type: (object) -> bool
        """
        Compare with another object for identity.

        :param other: Another object.

        :return: True if the same object.
        :rtype: bool
        """
        return other is self

    @final
    @recursive_repr
    def __repr__(self):
        # type: () -> str
        """
        Get representation.

        :return: Representation.
        :rtype: str
        """
        return custom_mapping_repr(
            self.to_dict(),
            prefix="{}(".format(type(self).__name__),
            template="{key}={value}",
            suffix=")",
            key_repr=str,
        )

    def to_dict(self):
        # type: () -> Dict[str, Any]
        """
        Convert to dictionary.

        :return: Dictionary.
        :rtype: dict[str, Any]
        """
        return {
            "relationship": self.relationship,
            "default": self.default,
            "default_factory": import_factory(self.default_factory),
            "module": self.module,
            "required": self.required,
            "changeable": self.changeable,
            "deletable": self.deletable,
            "finalized": self.finalized,
            "abstracted": self.abstracted,
        }

    def get_name(self, instance):
        # type: (BaseAttributeStructure) -> str
        """
        Get attribute name.

        :param instance: Instance.
        :type instance: objetto.bases.BaseAttributeStructure

        :return: Attribute Name.
        :rtype: str
        """
        cls = type(instance)
        return cls._attribute_names[self]

    def get_value(self, instance):
        # type: (BaseAttributeStructure) -> T
        """
        Get attribute value.

        :param instance: Instance.
        :type instance: objetto.bases.BaseAttributeStructure

        :return: Attribute Value.

        :raises AttributeError: Could not get value.
        """
        name = self.get_name(instance)
        try:
            value = instance[name]
        except KeyError:
            error = "attribute '{}' of '{}' has no value set".format(
                name, type(instance).__fullname__
            )
            if instance._initializing:
                error += " (instance hasn't finished initializing)"
            exc = AttributeError(error)
            raise_from(exc, None)
            raise exc
        else:
            return value

    @final
    def fabricate_default_value(self, **kwargs):
        # type: (Any, Any) -> Any
        """
        Fabricate default value.

        :param kwargs: Keyword arguments to be passed to the factories.

        :return: Fabricated value.

        :raises ValueError: No default or default factory.
        """
        default = self.__default
        if self.__default_factory is not None:
            default = run_factory(self.__default_factory, kwargs=kwargs)
        if default is MISSING:
            error = "attribute has no valid 'default' or 'default factory'"
            raise ValueError(error)
        return self.__relationship.fabricate_value(default, **kwargs)

    @property
    def relationship(self):
        # type: () -> BaseRelationship
        """
        Relationship.

        :rtype: objetto.bases.BaseRelationship
        """
        return self.__relationship

    @property
    def default(self):
        # type: () -> Any
        """Default value."""
        return self.__default

    @property
    def default_factory(self):
        # type: () -> LazyFactory
        """
        Default value factory.

        :rtype: str or collections.abc.Callable or None
        """
        return self.__default_factory

    @property
    def module(self):
        # type: () -> Optional[str]
        """
        Optional module path to use in case partial paths are provided.

        :rtype: str or None
        """
        return self.__module

    @property
    def required(self):
        # type: () -> bool
        """
        Whether attribute is required to have a value or not.

        :rtype: bool
        """
        return self.__required

    @property
    def changeable(self):
        # type: () -> bool
        """
        Whether attribute value can be changed.

        :rtype: bool
        """
        return self._changeable

    @property
    def deletable(self):
        # type: () -> bool
        """
        Whether attribute value can be deleted.

        :rtype: bool
        """
        return self._deletable

    @property
    def finalized(self):
        # type: () -> bool
        """
        If True, attribute can't be overridden by subclasses.

        :rtype: bool
        """
        return self.__finalized

    @property
    def abstracted(self):
        # type: () -> bool
        """
        If True, attribute needs to be overridden by subclasses.

        :rtype: bool
        """
        return self.__abstracted

    @property
    def has_default(self):
        # type: () -> bool
        """
        Whether attribute has a default value or a default factory.

        :rtype: bool
        """
        return self.default is not MISSING or self.default_factory is not None


class BaseAttributeStructureMeta(BaseStructureMeta):
    """Metaclass for `BaseAttributeStructure`."""

    __attributes = WeakKeyDictionary(
        {}
    )  # type: MutableMapping[BaseAttributeStructureMeta, Mapping[str, BaseAttribute]]
    __attribute_names = WeakKeyDictionary(
        {}
    )  # type: MutableMapping[BaseAttributeStructureMeta, Mapping[BaseAttribute, str]]

    def __init__(cls, name, bases, dct):
        # type: (str, Tuple[Type, ...], Dict[str, Any]) -> None
        super(BaseAttributeStructureMeta, cls).__init__(name, bases, dct)

        # Get attribute type.
        try:
            attribute_type = cls._attribute_type  # type: Optional[Type[BaseAttribute]]
        except NotImplementedError:
            attribute_type = None

        # Store attributes.
        attributes = {}
        if attribute_type is not None:
            for base in reversed(getmro(cls)):
                for member_name, member in iteritems(base.__dict__):
                    if isinstance(member, attribute_type):
                        attributes[member_name] = member
                    elif member_name in attributes:
                        del attributes[member_name]

        # Store attribute names.
        attribute_names = {}
        if attribute_type is not None:
            for attribute_name, attribute in iteritems(attributes):
                attribute_names[attribute] = attribute_name

        type(cls).__attributes[cls] = DictState(attributes)
        type(cls).__attribute_names[cls] = DictState(attribute_names)

    @property
    @abstractmethod
    def _attribute_type(cls):
        # type: () -> Type[BaseAttribute]
        """Attribute type."""
        raise NotImplementedError()

    @property
    def _attributes(cls):
        # type: () -> Mapping[str, BaseAttribute]
        """Attributes mapped by name."""
        return type(cls).__attributes[cls]

    @property
    def _attribute_names(cls):
        # type: () -> Mapping[Any, str]
        """Names mapped by attribute."""
        return type(cls).__attribute_names[cls]


# noinspection PyTypeChecker
_BAS = TypeVar("_BAS", bound="BaseAttributeStructure")


class BaseAttributeStructure(
    with_metaclass(BaseAttributeStructureMeta, BaseStructure[str])
):
    """
    Base attribute structure.

    Inherits from:
      - :class:`objetto.bases.BaseStructure`

    Inherited by:
      - :class:`objetto.data.Data`
      - :class:`objetto.objects.Object`

    Features:
      - Holds values in attributes defined by descriptors.
      - Can be cast into a dictionary.
    """

    __slots__ = ()

    @recursive_repr
    def __repr__(self):
        # type: () -> str
        """
        Get representation.

        :return: Representation.
        :rtype: str
        """

        class SimplifiedAuxiliary(object):
            __slots__ = ("auxiliary",)

            def __init__(self, auxiliary):
                self.auxiliary = auxiliary

            # noinspection PyCallingNonCallable
            def __repr__(self):
                if isinstance(self.auxiliary, BaseDict):
                    return recursive_repr(custom_mapping_repr)(self.auxiliary)
                elif isinstance(self.auxiliary, BaseList) or not self.auxiliary:
                    return recursive_repr(custom_iterable_repr)(self.auxiliary)
                else:
                    return recursive_repr(custom_iterable_repr)(
                        self.auxiliary, prefix="{", suffix="}"
                    )

        def simplify_auxiliary(value):
            if isinstance(value, (BaseDict, BaseList, BaseSet)):
                return SimplifiedAuxiliary(value)
            else:
                return value

        return custom_mapping_repr(
            dict(
                (n, simplify_auxiliary(v))
                for n, v in iteritems(self._state)
                if type(self)._get_relationship(n).represented
            ),
            prefix="{}(".format(type(self).__fullname__),
            template="{key}={value}",
            suffix=")",
            key_repr=str,
        )

    @final
    def __reversed__(self):
        # type: () -> Iterator[str]
        """
        Iterate over reversed attribute names.

        :return: Reversed attribute names iterator.
        :rtype: collections.abc.Iterator[str]
        """
        return self._state.__reversed__()

    @final
    def __getitem__(self, name):
        # type: (str) -> Any
        """
        Get value for attribute name.

        :param name: Attribute name.
        :type name: str

        :return: Value.

        :raises KeyError: Attribute does not exist or has no value.
        """
        return self._state[name]

    @final
    def __len__(self):
        # type: () -> int
        """
        Get count of attributes with value.

        :return: Count of attributes with value.
        :rtype: int
        """
        return len(self._state)

    @final
    def __iter__(self):
        # type: () -> Iterator[str]
        """
        Iterate over names of attributes with value.

        :return: Names of attributes with value.
        :rtype: collections.abc.Iterator[str]
        """
        for name in self._state:
            yield name

    @final
    def __contains__(self, name):
        # type: (Any) -> bool
        """
        Get whether attribute name is valid and has a value.

        :param name: Attribute name.
        :type name: str

        :return: True if attribute name is valid and has a value.
        :rtype: bool
        """
        return name in self._state

    @classmethod
    def _get_relationship(cls, location):
        # type: (str) -> BaseRelationship
        """
        Get relationship at location (attribute name).

        :param location: Location (attribute name).
        :type location: str

        :return: Relationship.
        :rtype: objetto.bases.BaseRelationship

        :raises KeyError: Attribute does not exist.
        """
        return cls._get_attribute(location).relationship

    @classmethod
    def _get_attribute(cls, name):
        # type: (str) -> BaseAttribute
        """
        Get attribute by name.

        :param name: Attribute name.
        :type name: str

        :return: Attribute.
        :rtype: objetto.bases.BaseAttribute

        :raises KeyError: Attribute does not exist.
        """
        return cls._attributes[name]

    @abstractmethod
    def _set(self, name, value):
        # type: (_BAS, str, Any) -> _BAS
        """
        Set attribute value.

        :param name: Attribute name.
        :type name: str

        :param value: Value.

        :return: Transformed.
        :rtype: objetto.bases.BaseAttributeStructure

        :raises AttributeError: Attribute is not changeable and already has a value.
        """
        raise NotImplementedError()

    @abstractmethod
    def _delete(self, name):
        # type: (_BAS, str) -> _BAS
        """
        Delete attribute value.

        :param name: Attribute name.
        :type name: str

        :return: Transformed.
        :rtype: objetto.bases.BaseAttributeStructure

        :raises KeyError: Attribute does not exist or has no value.
        :raises AttributeError: Attribute is not deletable.
        """
        raise NotImplementedError()

    @overload
    @abstractmethod
    def _update(self, __m, **kwargs):
        # type: (_BAS, Mapping[str, Any], Any) -> _BAS
        pass

    @overload
    @abstractmethod
    def _update(self, __m, **kwargs):
        # type: (_BAS, Iterable[Tuple[str, Any]], Any) -> _BAS
        pass

    @overload
    @abstractmethod
    def _update(self, **kwargs):
        # type: (_BAS, Any) -> _BAS
        pass

    @abstractmethod
    def _update(self, *args, **kwargs):
        """
        Update multiple attribute values.
        Same parameters as :meth:`dict.update`.

        :return: Transformed.
        :rtype: objetto.bases.BaseAttributeStructure

        :raises AttributeError: Attribute is not changeable and already has a value.
        """
        raise NotImplementedError()

    @final
    def keys(self):
        # type: () -> SetState[str]
        """
        Get names of the attributes with values.

        :return: Attribute names.
        :rtype: objetto.states.SetState[str]
        """
        return SetState(self._state.keys())

    @final
    def find_with_attributes(self, **attributes):
        # type: (Any) -> Any
        """
        Find first value that matches unique attribute values.

        :param attributes: Attributes to match.

        :return: Value.

        :raises ValueError: No attributes provided or no match found.
        """
        return self._state.find_with_attributes(**attributes)

    @property
    @abstractmethod
    def _state(self):
        # type: () -> DictState[str, Any]
        """
        Internal state.

        :rtype: objetto.states.DictState[str, Any]
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class BaseInteractiveAttributeStructure(
    BaseAttributeStructure, BaseInteractiveStructure[str]
):
    """
    Base interactive attribute structure.

    Inherits from:
      - :class:`objetto.bases.BaseAttributeStructure`
      - :class:`objetto.bases.BaseInteractiveStructure`

    Inherited by:
      - :class:`objetto.data.InteractiveData`
    """

    __slots__ = ()

    @final
    def set(self, name, value):
        # type: (_BAS, str, Any) -> _BAS
        """
        Set attribute value.

        :param name: Attribute name.
        :type name: str

        :param value: Value.

        :return: Transformed.
        :rtype: objetto.bases.BaseInteractiveAttributeStructure

        :raises AttributeError: Attribute is not changeable and already has a value.
        """
        return self._set(name, value)

    @final
    def delete(self, name):
        # type: (_BAS, str) -> _BAS
        """
        Delete attribute value.

        :param name: Attribute name.
        :type name: str

        :return: Transformed.
        :rtype: objetto.bases.BaseInteractiveAttributeStructure

        :raises KeyError: Attribute does not exist or has no value.
        :raises AttributeError: Attribute is not deletable.
        """
        return self._delete(name)

    @overload
    def update(self, __m, **kwargs):
        # type: (_BAS, Mapping[str, Any], Any) -> _BAS
        pass

    @overload
    def update(self, __m, **kwargs):
        # type: (_BAS, Iterable[Tuple[str, Any]], Any) -> _BAS
        pass

    @overload
    def update(self, **kwargs):
        # type: (_BAS, Any) -> _BAS
        pass

    @final
    def update(self, *args, **kwargs):
        """
        Update multiple attribute values.
        Same parameters as :meth:`dict.update`.

        :return: Transformed.
        :rtype: objetto.bases.BaseInteractiveAttributeStructure

        :raises AttributeError: Attribute is not changeable and already has a value.
        """
        return self._update(*args, **kwargs)


# noinspection PyAbstractClass
class BaseMutableAttributeStructure(BaseAttributeStructure, BaseMutableStructure[str]):
    """
    Base mutable attribute structure.

    Inherits from:
      - :class:`objetto.bases.BaseAttributeStructure`
      - :class:`objetto.bases.BaseMutableStructure`

    Inherited by:
      - :class:`objetto.objects.Object`
    """

    __slots__ = ()

    @final
    def __setitem__(self, name, value):
        # type: (str, Any) -> None
        """
        Set attribute value.

        :param name: Attribute name.
        :type name: str

        :return: Value.

        :raises KeyError: Attribute does not exist.
        """
        self._set(name, value)

    @final
    def __delitem__(self, name):
        # type: (str) -> None
        """
        Delete attribute value.

        :param name: Attribute name.
        :type name: str

        :raises KeyError: Attribute does not exist or has no value.
        """
        self._delete(name)

    @final
    def delete(self, name):
        # type: (str) -> None
        """
        Delete attribute value.

        :param name: Attribute name.
        :type name: str

        :raises KeyError: Attribute does not exist or has no value.
        """
        self._delete(name)

    @final
    def set(self, name, value):
        # type: (str, Any) -> None
        """
        Set attribute value.

        :param name: Attribute name.
        :type name: str

        :param value: Value.

        :raises KeyError: Attribute does not exist.
        """
        self._set(name, value)

    @overload
    def update(self, __m, **kwargs):
        # type: (Mapping[str, Any], Any) -> None
        pass

    @overload
    def update(self, __m, **kwargs):
        # type: (Iterable[Tuple[str, Any]], Any) -> None
        pass

    @overload
    def update(self, **kwargs):
        # type: (Any) -> None
        pass

    @final
    def update(self, *args, **kwargs):
        """
        Update multiple attribute values.
        Same parameters as :meth:`dict.update`.

        :raises KeyError: Attribute does not exist.
        """
        self._update(*args, **kwargs)



# -*- coding: utf-8 -*-
"""Generate importable lazy paths and import from them."""

from re import match as re_match
from typing import TYPE_CHECKING

from qualname import qualname  # type: ignore

if TYPE_CHECKING:
    from typing import Any, Optional

__all__ = [
    "PARTIAL_IMPORT_PATH_REGEX",
    "RELATIVE_IMPORT_PATH_REGEX",
    "IMPORT_PATH_REGEX",
    "import_path",
    "get_path",
    "decorate_path",
]


PARTIAL_IMPORT_PATH_REGEX = r"^((?:\w+\.{0,1})+)$"
"""Partial lazy import path regex."""

RELATIVE_IMPORT_PATH_REGEX = (
    r"^((?:(?:\.+)|(?:\.*)(?:\w+\.{0,1})+))\|((?:\w+\.{0,1})+)$"
)
"""Relative lazy import path regex."""

IMPORT_PATH_REGEX = r"^((?:\w+\.{0,1})+)\|((?:\w+\.{0,1})+)$"
"""Full lazy import path regex."""


def import_path(path):
    # type: (str) -> Any
    """
    Import object from a full import path.

    .. code:: python

        >>> from objetto.utils.lazy_import import import_path

        >>> import_path("abc|abstractmethod")
        <function abstractmethod at ...>

    :param path: Import path.
    :type path: str

    :return: Imported object.

    :raises ValueError: Invalid or empty path.
    :raises AttributeError: No object with the provided name.
    """
    if "|" not in path:
        if not path:
            error = "can't import from empty path"
        else:
            error = (
                "import path '{}' does not specify a module name (missing '|' "
                "separator between module path and qualified name)"
            ).format(path)
        raise ValueError(error)
    elif path.startswith("."):
        error = "import path '{}' is not absolute".format(path)
        raise ValueError(error)

    match = re_match(IMPORT_PATH_REGEX, path)
    if not match:
        error = "invalid import path '{}'".format(path)
        raise ValueError(error)

    module, qual_name = match.groups()
    name_parts = qual_name.split(".")
    module_obj = __import__(module, fromlist=[name_parts[0]])

    obj = module_obj
    for name_part in name_parts:
        obj = getattr(obj, name_part)

    return obj


def get_path(obj):
    # type: (Any) -> str
    """
    Get full import path to an object.

    .. code:: python

        >>> from abc import abstractmethod
        >>> from objetto.utils.lazy_import import get_path

        >>> get_path(abstractmethod)
        'abc|abstractmethod'

    :param obj: Object.

    :return: Import path.
    :rtype: str

    :raises ValueError: Can't determine consistent import path.
    """
    module = obj.__module__
    if not module:
        error = "can't get module for {}".format(obj)
        raise ValueError(error)

    try:
        qual_name = qualname(obj)
        if not qual_name:
            raise AttributeError()
    except AttributeError:
        qual_name = obj.__name__

    if not qual_name:
        error = "can't get name for {}".format(obj)
        raise ValueError(error)

    path = "|".join((module, qual_name))

    try:
        imported_obj = import_path(path)
    except ImportError:
        imported_obj = None

    if imported_obj is not obj:
        error = "can't get consistent import path to {}".format(obj)
        raise ValueError(error)

    return path


def decorate_path(path, module=None):
    # type: (str, Optional[str]) -> str
    """
    Validate and decorate partial/relative import path with module if applicable.

    .. code:: python

        >>> from objetto.utils.lazy_import import decorate_path

        >>> decorate_path("abstractmethod", module="abc")
        'abc|abstractmethod'

    :param path: Import path (partial or full)
    :type path: str

    :param module: Optional module path.
    :type module: str or None

    :return: Full import path.
    :rtype: str

    :raises ValueError: Invalid path or no module provided.
    """
    if not path:
        error = "can't decorate empty path"
        raise ValueError(error)

    # Partial path, need to add module.
    if "|" not in path:
        path_match = re_match(PARTIAL_IMPORT_PATH_REGEX, path)
        if not path_match:
            error = "invalid partial import path '{}'".format(path)
            raise ValueError(error)

        if not module:
            error = (
                "can't add module to partial path '{}' since it was not provided"
            ).format(path)
            raise ValueError(error)

        module_match = re_match(PARTIAL_IMPORT_PATH_REGEX, module)
        if not module_match:
            error = "invalid module path '{}'".format(module)
            raise ValueError(error)

        return module + "|" + path

    # Relative path, need to resolve module and decorate it.
    elif path.startswith("."):
        path_match = re_match(RELATIVE_IMPORT_PATH_REGEX, path)
        if not path_match:
            error = "invalid relative import path '{}'".format(path)
            raise ValueError(error)

        if not module:
            error = (
                "can't resolve module to relative path '{}' since it was not provided"
            ).format(path)
            raise ValueError(error)

        module_match = re_match(PARTIAL_IMPORT_PATH_REGEX, module)
        if not module_match:
            error = "invalid module path '{}'".format(module)
            raise ValueError(error)

        path = path[1:]
        if path.startswith("|"):
            return module + path
        elif not path.startswith("."):
            path_parts = path.split("|")
            return module + "." + path_parts[0] + "|" + path_parts[1]
        else:
            while path.startswith("."):
                module_parts = module.split(".")
                if len(module_parts) < 2:
                    error = "can't get parent module for '{}'".format(module)
                    raise ValueError(error)
                module, path = ".".join(module_parts[:-1]), path[1:]
            if "|" in path and not path.startswith("|"):
                module, path = (
                    module + "." + path.split("|")[0],
                    "|" + path.split("|")[1],
                )

        return module + path

    # Full path, no need to decorate it with module.
    else:
        match = re_match(IMPORT_PATH_REGEX, path)
        if not match:
            error = "invalid import path '{}'".format(path)
            raise ValueError(error)

        return path



# -*- coding: utf-8 -*-
"""Immutable state types."""

from abc import abstractmethod
from typing import TYPE_CHECKING, TypeVar, cast

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

from .._bases import BaseHashable, BaseInteractiveCollection, final

if TYPE_CHECKING:
    from typing import Any, Optional, Type, Union

    from pyrsistent.typing import PMap, PSet, PVector

__all__ = ["BaseState"]


T = TypeVar("T")  # Any type.
KT = TypeVar("KT")  # Key type.
VT = TypeVar("VT")  # Value type.


# noinspection PyTypeChecker
_BS = TypeVar("_BS", bound="BaseState")


class BaseState(BaseHashable, BaseInteractiveCollection[T]):
    """
    Base immutable state.

    Inherits from:
      - :class:`objetto.bases.BaseHashable`
      - :class:`objetto.bases.BaseInteractiveCollection`

    Inherited By:
      - :class:`objetto.states.DictState`
      - :class:`objetto.states.ListState`
      - :class:`objetto.states.SetState`

    :param initial: Initial values.

    :raises NotImplementedError: Abstract methods not implemented.
    """

    __slots__ = ("__hash", "__internal")

    @classmethod
    @abstractmethod
    def _make(cls, internal):
        # type: (Type[_BS], Any) -> _BS
        """
        Make new state by directly setting the internal state.

        :param internal: Internal state.

        :return: State.
        :rtype: objetto.bases.BaseState
        """
        self = cast("_BS", cls.__new__(cls))
        self.__internal = internal
        self.__hash = None
        return self

    @staticmethod
    @abstractmethod
    def _make_internal(initial):
        # type: (Any) -> Union[PMap, PVector, PSet]
        """
        Initialize internal state.

        :param initial: Initial values.
        """
        raise NotImplementedError()

    @abstractmethod
    def __init__(self, initial=()):
        # type: (Any) -> None
        self.__internal = self._make_internal(initial)
        self.__hash = None  # type: Optional[int]

    def __hash__(self):
        # type: () -> int
        """
        Get hash.

        :return: Hash.
        :rtype: int
        """
        if self.__hash is None:
            self.__hash = hash(self._internal)
        return self.__hash

    @abstractmethod
    def __eq__(self, other):
        # type: (object) -> bool
        """
        Compare for equality.

        :param other: Another object.

        :return: True if equal.
        :rtype: bool

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @final
    def __copy__(self):
        # type: (_BS) -> _BS
        """
        Get copy.

        :return: Copy.
        :rtype: objetto.bases.BaseState
        """
        return self

    @abstractmethod
    def __repr__(self):
        # type: () -> str
        """
        Get representation.

        :return: Representation.
        :rtype: str

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    def __str__(self):
        # type: () -> str
        """
        Get string representation.

        :return: String representation.
        :rtype: str
        """
        return self.__repr__()

    @property
    @abstractmethod
    def _internal(self):
        # type: () -> Union[PMap, PVector, PSet]
        """Internal values."""
        return self.__internal



# -*- coding: utf-8 -*-
"""Data with state curated by attribute descriptors."""

from typing import TYPE_CHECKING, TypeVar, cast, overload

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

from six import iteritems, string_types, with_metaclass

from .._bases import MISSING, final
from .._states import BaseState, DictState
from .._structures import (
    BaseAttribute,
    BaseAttributeMeta,
    BaseAttributeStructure,
    BaseAttributeStructureMeta,
    BaseInteractiveAttributeStructure,
)
from ..utils.type_checking import assert_is_instance
from .bases import BaseData, BaseDataMeta, BaseInteractiveData, DataRelationship

if TYPE_CHECKING:
    from typing import Any, Dict, Iterable, Mapping, Optional, Set, Tuple, Type

    from ..utils.factoring import LazyFactory

__all__ = ["DataAttribute", "Data", "InteractiveData"]


T = TypeVar("T")  # Any type.


class DataAttributeMeta(BaseAttributeMeta):
    """Metaclass for `DataAttribute`."""

    @property
    @final
    def _relationship_type(cls):
        # type: () -> Type[DataRelationship]
        """Relationship type."""
        return DataRelationship


@final
class DataAttribute(with_metaclass(DataAttributeMeta, BaseAttribute[T])):
    """
    Attribute descriptor for :class:`objetto.data.Data` classes.

    Inherits from:
      - :class:`objetto.bases.BaseAttribute`

    :param relationship: Relationship.
    :type relationship: objetto.data.DataRelationship

    :param default: Default value.

    :param default_factory: Default value factory.
    :type default_factory: str or collections.abc.Callable or None

    :param module: Optional module path to use in case partial paths are provided.
    :type module: str or None

    :param required: Whether attribute is required to have a value or not.
    :type required: bool

    :param changeable: Whether attribute value can be changed.
    :type changeable: bool

    :param deletable: Whether attribute value can be deleted.
    :type deletable: bool

    :param finalized: If True, attribute can't be overridden by subclasses.
    :type finalized: bool

    :param abstracted: If True, attribute needs to be overridden by subclasses.
    :type abstracted: bool

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    """

    __slots__ = ()

    def __init__(
        self,
        relationship=DataRelationship(),  # type: DataRelationship
        default=MISSING,  # type: Any
        default_factory=None,  # type: LazyFactory
        module=None,  # type: Optional[str]
        required=True,  # type: bool
        changeable=True,  # type: bool
        deletable=False,  # type: bool
        finalized=False,  # type: bool
        abstracted=False,  # type: bool
    ):
        # type: (...) -> None
        super(DataAttribute, self).__init__(
            relationship=relationship,
            default=default,
            default_factory=default_factory,
            module=module,
            required=required,
            changeable=changeable,
            deletable=deletable,
            finalized=finalized,
            abstracted=abstracted,
        )

    @property
    def relationship(self):
        # type: () -> DataRelationship
        """
        Relationship.

        :rtype: objetto.data.DataRelationship
        """
        return cast("DataRelationship", super(DataAttribute, self).relationship)


class DataMeta(BaseAttributeStructureMeta, BaseDataMeta):
    """Metaclass for `Data`."""

    @property
    @final
    def _attribute_type(cls):
        # type: () -> Type[DataAttribute]
        """Attribute type."""
        return DataAttribute

    @property
    @final
    def _attributes(cls):
        # type: () -> Mapping[str, DataAttribute]
        """Attributes mapped by name."""
        return cast("Mapping[str, DataAttribute]", super(DataMeta, cls)._attributes)

    @property
    @final
    def _attribute_names(cls):
        # type: () -> Mapping[DataAttribute, str]
        """Names mapped by attribute."""
        return cast(
            "Mapping[DataAttribute, str]", super(DataMeta, cls)._attribute_names
        )


# noinspection PyTypeChecker
_D = TypeVar("_D", bound="Data")


class Data(with_metaclass(DataMeta, BaseAttributeStructure, BaseData[str])):
    """
    Data.

    Inherits from:
      - :class:`objetto.bases.BaseAttributeStructure`
      - :class:`objetto.bases.BaseData`

    Inherited by:
      - :class:`objetto.data.InteractiveData`
      - :class:`objetto.bases.BaseChange`
      - :class:`objetto.objects.Action`
      - :class:`objetto.observers.ActionObserverExceptionData`

    :param initial: Initial values.
    """

    __slots__ = ("__hash",)

    @classmethod
    @final
    def __make__(cls, state=DictState()):
        # type: (Type[_D], BaseState) -> _D
        """
        Make a new data.

        :param state: Internal state.
        :return: New data.
        """
        return super(Data, cls).__make__(state)

    @final
    def __init__(self, **initial):
        # type: (Any) -> None
        state = self.__get_initial_state(initial)
        self._init_state(state)

    @classmethod
    @final
    def __get_initial_state(
        cls,
        input_values,  # type: Mapping[str, Any]
        factory=True,  # type: bool
    ):
        # type: (...) -> DictState[str, Any]
        """
        Get initial state.

        :param input_values: Input values.
        :param factory: Whether to run values through factory.
        :return: Initial state.
        """
        initial = {}

        for name, value in iteritems(input_values):
            attribute = cls._get_attribute(name)
            initial[name] = attribute.relationship.fabricate_value(
                value, factory=factory
            )

        missing_attributes = set()  # type: Set[str]
        for name, attribute in iteritems(cls._attributes):
            if name not in initial:
                if attribute.has_default:
                    if not missing_attributes:
                        initial[name] = attribute.fabricate_default_value()
                elif attribute.required:
                    missing_attributes.add(name)

        if missing_attributes:
            error = "missing required attribute{} {}".format(
                "s" if len(missing_attributes) != 1 else "",
                ", ".join("'{}'".format(n) for n in missing_attributes),
            )
            raise TypeError(error)

        return DictState(initial)

    @classmethod
    @final
    def _get_relationship(cls, location):
        # type: (str) -> DataRelationship
        """
        Get relationship at location (attribute name).

        :param location: Location (attribute name).
        :type location: str

        :return: Relationship.
        :rtype: objetto.data.DataRelationship

        :raises KeyError: Attribute does not exist.
        """
        return cast("DataRelationship", cls._get_attribute(location).relationship)

    @classmethod
    @final
    def _get_attribute(cls, name):
        # type: (str) -> DataAttribute
        """
        Get attribute by name.

        :param name: Attribute name.
        :type name: str

        :return: Attribute.
        :rtype: objetto.data.DataAttribute

        :raises KeyError: Attribute does not exist.
        """
        return cast("DataAttribute", cls._attributes[name])

    @final
    def _hash(self):
        # type: () -> int
        """
        Get hash.

        :return: Hash.
        :rtype: int
        """
        try:
            return self.__hash  # type: ignore
        except AttributeError:
            cls = type(self)
            comparable_attributes = set(
                n for n, a in iteritems(cls._attributes) if a.relationship.compared
            )
            if not comparable_attributes:
                return hash(id(self))
            comparable_state = dict(
                (n, v) for n, v in iteritems(self._state) if n in comparable_attributes
            )
            self.__hash = hash(frozenset(iteritems(comparable_state)))
            return self.__hash

    @final
    def _eq(self, other):
        # type: (Any) -> bool
        """
        Compare with another data for equality.

        :param other: Another data.

        :return: True if equal.
        :rtype: bool
        """
        if self is other:
            return True
        cls = type(self)
        comparable_attributes = set(
            n for n, a in iteritems(cls._attributes) if a.relationship.compared
        )
        if not comparable_attributes:
            return False
        comparable_state = dict(
            (n, v)
            for n, v in iteritems(self._state)
            if cls._get_relationship(n).compared
        )
        if not isinstance(other, collections_abc.Hashable):
            return comparable_state == other
        if not isinstance(other, Data):
            return False
        if type(self) is not type(other):
            return False
        other_comparable_state = dict(
            (n, v)
            for n, v in iteritems(other._state)
            if cls._get_relationship(n).compared
        )
        return comparable_state == other_comparable_state

    @final
    def _clear(self):
        # type: (_D) -> _D
        """
        Clear deletable attribute values.

        :return: Transformed.
        :rtype: objetto.data.Data

        :raises AttributeError: No deletable attributes.
        """
        cls = type(self)
        state = self._state
        has_deletable_attributes = False
        for name in self._state:
            attribute = cls._get_attribute(name)
            if attribute.deletable:
                has_deletable_attributes = True
                state = state.remove(name)
        if not has_deletable_attributes:
            error = "'{}' has no deletable attributes".format(type(self).__fullname__)
            raise AttributeError(error)
        if state is self._state:
            return self
        else:
            return type(self).__make__(state)

    @final
    def _set(self, name, value):
        # type: (_D, str, Any) -> _D
        """
        Set attribute value.

        :param name: Attribute name.
        :type name: str

        :param value: Value.

        :return: Transformed.
        :rtype: objetto.data.Data

        :raises AttributeError: Attribute is not changeable and already has a value.
        """
        cls = type(self)
        attribute = cls._get_attribute(name)
        if not attribute.changeable and name in self._state:
            error = "non-changeable attribute '{}' already has a value".format(name)
            raise AttributeError(error)
        fabricated_value = attribute.relationship.fabricate_value(value)
        return type(self).__make__(self._state.set(name, fabricated_value))

    @final
    def _delete(self, name):
        # type: (_D, str) -> _D
        """
        Delete attribute value.

        :param name: Attribute name.
        :type name: str

        :return: Transformed.
        :rtype: objetto.data.Data

        :raises KeyError: Attribute does not exist or has no value.
        :raises AttributeError: Attribute is not deletable.
        """
        cls = type(self)
        attribute = cls._get_attribute(name)
        if not attribute.deletable and name in self._state:
            error = "attribute '{}' is not deletable".format(name)
            raise AttributeError(error)
        return type(self).__make__(self._state.remove(name))

    @overload
    def _update(self, __m, **kwargs):
        # type: (_D, Mapping[str, Any], Any) -> _D
        pass

    @overload
    def _update(self, __m, **kwargs):
        # type: (_D, Iterable[Tuple[str, Any]], Any) -> _D
        pass

    @overload
    def _update(self, **kwargs):
        # type: (_D, Any) -> _D
        pass

    @final
    def _update(self, *args, **kwargs):
        """
        Update multiple attribute values.
        Same parameters as :meth:`dict.update`.

        :return: Transformed.
        :rtype: objetto.data.Data

        :raises AttributeError: Attribute is not changeable and already has a value.
        """
        update = dict(*args, **kwargs)
        cls = type(self)
        fabricated_update = {}
        for name, value in iteritems(update):
            assert_is_instance(name, string_types)
            attribute = cls._get_attribute(name)
            if not attribute.changeable and name in self._state:
                error = "non-changeable attribute '{}' already has a value".format(name)
                raise AttributeError(error)
            fabricated_value = attribute.relationship.fabricate_value(value)
            fabricated_update[name] = fabricated_value
        return cls.__make__(self._state.update(fabricated_update))

    @classmethod
    @final
    def deserialize(cls, serialized, **kwargs):
        # type: (Type[_D], Dict, Any) -> _D
        """
        Deserialize.

        :param serialized: Serialized.

        :param kwargs: Keyword arguments to be passed to the deserializers.

        :return: Deserialized.
        :rtype: objetto.data.Data
        """
        input_values = dict(
            (n, cls.deserialize_value(v, n, **kwargs)) for n, v in iteritems(serialized)
        )
        state = cls.__get_initial_state(input_values, factory=False)
        return cls.__make__(state)

    @final
    def serialize(self, **kwargs):
        # type: (Any) -> Dict[str, Any]
        """
        Serialize.

        :param kwargs: Keyword arguments to be passed to the serializers.

        :return: Serialized.
        :rtype: dict[str, Any]
        """
        return dict(
            (n, self.serialize_value(v, n, **kwargs))
            for n, v in iteritems(self._state)
            if type(self)._get_relationship(n).serialized
        )

    @property
    @final
    def _state(self):
        # type: () -> DictState[str, Any]
        """
        Internal state.

        :rtype: objetto.states.DictState[str, Any]
        """
        return cast("DictState", super(BaseAttributeStructure, self)._state)


class InteractiveData(
    Data, BaseInteractiveAttributeStructure, BaseInteractiveData[str]
):
    """
    Interactive data.

    Inherits from:
      - :class:`objetto.data.Data`
      - :class:`objetto.bases.BaseInteractiveAttributeStructure`
      - :class:`objetto.bases.BaseInteractiveData`
    """

    __slots__ = ()



# -*- coding: utf-8 -*-
"""State-carrying structures."""

from .bases import (
    BaseAuxiliaryStructure,
    BaseAuxiliaryStructureMeta,
    BaseInteractiveAuxiliaryStructure,
    BaseInteractiveStructure,
    BaseMutableAuxiliaryStructure,
    BaseMutableStructure,
    BaseRelationship,
    BaseStructure,
    BaseStructureMeta,
    SerializationError,
    UniqueDescriptor,
    make_auxiliary_cls,
    unique_descriptor,
)
from .dict import (
    BaseDictStructure,
    BaseDictStructureMeta,
    BaseInteractiveDictStructure,
    BaseMutableDictStructure,
    KeyRelationship,
)
from .list import (
    BaseInteractiveListStructure,
    BaseListStructure,
    BaseListStructureMeta,
    BaseMutableListStructure,
)
from .set import (
    BaseInteractiveSetStructure,
    BaseMutableSetStructure,
    BaseSetStructure,
    BaseSetStructureMeta,
)
from .structure import (
    BaseAttribute,
    BaseAttributeMeta,
    BaseAttributeStructure,
    BaseAttributeStructureMeta,
    BaseInteractiveAttributeStructure,
    BaseMutableAttributeStructure,
)

__all__ = [
    "make_auxiliary_cls",
    "unique_descriptor",
    "UniqueDescriptor",
    "SerializationError",
    "BaseRelationship",
    "BaseStructureMeta",
    "BaseStructure",
    "BaseInteractiveStructure",
    "BaseMutableStructure",
    "BaseAttributeMeta",
    "BaseAttribute",
    "BaseAttributeStructureMeta",
    "BaseAttributeStructure",
    "BaseInteractiveAttributeStructure",
    "BaseMutableAttributeStructure",
    "BaseAuxiliaryStructureMeta",
    "BaseAuxiliaryStructure",
    "BaseInteractiveAuxiliaryStructure",
    "BaseMutableAuxiliaryStructure",
    "KeyRelationship",
    "BaseDictStructureMeta",
    "BaseDictStructure",
    "BaseInteractiveDictStructure",
    "BaseMutableDictStructure",
    "BaseListStructureMeta",
    "BaseListStructure",
    "BaseInteractiveListStructure",
    "BaseMutableListStructure",
    "BaseSetStructureMeta",
    "BaseSetStructure",
    "BaseInteractiveSetStructure",
    "BaseMutableSetStructure",
]



# -*- coding: utf-8 -*-
"""Immutable structures."""

from abc import abstractmethod
from typing import TYPE_CHECKING, TypeVar, cast

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

from six import with_metaclass

from .._bases import final, init_context
from .._states import BaseState
from .._structures import (
    BaseAuxiliaryStructure,
    BaseAuxiliaryStructureMeta,
    BaseInteractiveAuxiliaryStructure,
    BaseInteractiveStructure,
    BaseRelationship,
    BaseStructure,
    BaseStructureMeta,
)

if TYPE_CHECKING:
    from typing import Any, Dict, Optional, Tuple, Type

    from ..utils.factoring import LazyFactory
    from ..utils.type_checking import LazyTypes

__all__ = [
    "DataRelationship",
    "BaseDataMeta",
    "BaseData",
    "BaseInteractiveData",
]


T = TypeVar("T")  # Any type.


@final
class DataRelationship(BaseRelationship):
    """
    Relationship between a data structure and its values.

    Inherits from:
      - :class:`objetto.bases.BaseRelationship`

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :param checked: Whether to perform runtime type check.
    :type checked: bool

    :param module: Module path for lazy types/factories.
    :type module: str or None

    :param factory: Value factory.
    :type factory: str or collections.abc.Callable or None

    :param serialized: Whether should be serialized.
    :type serialized: bool

    :param serializer: Custom serializer.
    :type serializer: str or collections.abc.Callable or None

    :param deserializer: Custom deserializer.
    :type deserializer: str or collections.abc.Callable or None

    :param represented: Whether should be represented.
    :type represented: bool

    :param compared: Whether the value should be leverage when comparing.
    :type compared: bool

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    """

    __slots__ = ("__compared",)

    def __init__(
        self,
        types=(),  # type: LazyTypes
        subtypes=False,  # type: bool
        checked=None,  # type: Optional[bool]
        module=None,  # type: Optional[str]
        factory=None,  # type: LazyFactory
        serialized=True,  # type: bool
        serializer=None,  # type: LazyFactory
        deserializer=None,  # type: LazyFactory
        represented=True,  # type: bool
        compared=True,  # type: bool
    ):
        # type: (...) -> None
        super(DataRelationship, self).__init__(
            types=types,
            subtypes=subtypes,
            checked=checked,
            module=module,
            factory=factory,
            serialized=serialized,
            serializer=serializer,
            deserializer=deserializer,
            represented=represented,
        )
        self.__compared = bool(compared)

    def to_dict(self):
        # type: () -> Dict[str, Any]
        """
        Convert to dictionary.

        :return: Dictionary.
        :rtype: dict[str, Any]
        """
        dct = super(DataRelationship, self).to_dict()
        dct.update(
            {
                "compared": self.compared,
            }
        )
        return dct

    @property
    def compared(self):
        # type: () -> bool
        """
        Whether the value should be leverage when comparing.

        :rtype: bool
        """
        return self.__compared


class BaseDataMeta(BaseStructureMeta):
    """Metaclass for `BaseData`."""

    @property
    @final
    def _serializable_container_types(cls):
        # type: () -> Tuple[Type[BaseData]]
        """Serializable container types."""
        return (BaseData,)

    @property
    @final
    def _serializable_structure_types(cls):
        # type: () -> Tuple[Type[BaseData]]
        """Serializable structure types."""
        return (BaseData,)

    @property
    @final
    def _relationship_type(cls):
        # type: () -> Type[BaseRelationship]
        """Relationship type."""
        return DataRelationship


# noinspection PyTypeChecker
_BD = TypeVar("_BD", bound="BaseData")


# noinspection PyAbstractClass
class BaseData(with_metaclass(BaseDataMeta, BaseStructure[T])):
    """
    Base data.

    Inherits from:
      - :class:`objetto.bases.BaseStructure`

    Inherited by:
      - :class:`objetto.bases.BaseInteractiveData`
      - :class:`objetto.bases.BaseAuxiliaryData`
      - :class:`objetto.data.Data`

    Features:
      - Is an immutable protected structure.
    """

    __slots__ = ("__state",)

    @final
    def __copy__(self):
        # type: (_BD) -> _BD
        """
        Get copy.

        :return: Copy.
        :rtype: objetto.bases.BaseData
        """
        return self

    @classmethod
    def __make__(cls, state):
        # type: (Type[_BD], Any) -> _BD
        """
        Make a new data.

        :param state: Internal state.
        :type state: objetto.bases.BaseState

        :return: New data.
        :rtype: objetto.bases.BaseData
        """
        self = cast("_BD", cls.__new__(cls))
        self._init_state(state)
        return self

    @final
    def _init_state(self, state):
        # type: (BaseState) -> None
        """
        Initialize internal state.

        :param state: Internal state.
        :type state: objetto.bases.BaseState

        :raises RuntimeError: State already initialized.
        """
        try:
            _ = self.__state  # type: ignore
        except AttributeError:
            with init_context(self):
                self.__state = state
        else:
            error = "state already initialized"
            raise RuntimeError(error)

    @property
    def _state(self):
        # type: () -> BaseState
        """
        State.

        :rtype: objetto.bases.BaseState
        """
        return self.__state


# noinspection PyAbstractClass
class BaseInteractiveData(BaseData[T], BaseInteractiveStructure[T]):
    """
    Base interactive data.

    Inherits from:
      - :class:`objetto.bases.BaseData`
      - :class:`objetto.bases.BaseInteractiveStructure`

    Inherited by:
      - :class:`objetto.bases.BaseInteractiveAuxiliaryData`
      - :class:`objetto.data.InteractiveData`

    Features:
      - Is an immutable interactive data structure.
    """

    __slots__ = ()


class BaseAuxiliaryDataMeta(BaseDataMeta, BaseAuxiliaryStructureMeta):
    """Metaclass for `BaseAuxiliaryData`."""

    @property
    @abstractmethod
    def _base_auxiliary_type(cls):
        # type: () -> Type[BaseAuxiliaryData]
        """Base auxiliary data type."""
        raise NotImplementedError()


# noinspection PyAbstractClass
class BaseAuxiliaryData(
    with_metaclass(
        BaseAuxiliaryDataMeta,
        BaseAuxiliaryStructure[T],
        BaseData[T],
    )
):
    """
    Base auxiliary data.

    Inherits from:
      - :class:`objetto.bases.BaseAuxiliaryStructure`
      - :class:`objetto.bases.BaseData`

    Inherited by:
      - :class:`objetto.bases.BaseInteractiveAuxiliaryData`
      - :class:`objetto.data.InteractiveDictData`
      - :class:`objetto.data.InteractiveListData`
      - :class:`objetto.data.InteractiveSetData`
    """

    __slots__ = ("__hash",)

    _relationship = DataRelationship()
    """
    Relationship for all locations.

    :type: objetto.data.DataRelationship
    """

    @final
    def _hash(self):
        # type: () -> int
        """
        Get hash.

        :return: Hash.
        :rtype: int
        """
        try:
            return self.__hash  # type: ignore
        except AttributeError:
            if not type(self)._relationship.compared:
                self.__hash = hash(id(self))
            else:
                self.__hash = hash(self._state)
            return self.__hash

    @final
    def _eq(self, other):
        # type: (Any) -> bool
        """
        Compare with another object for equality.

        :param other: Another object.

        :return: True if equal.
        :rtype: bool
        """
        if self is other:
            return True
        self_compared = type(self)._relationship.compared
        if not self_compared:
            return False
        if not isinstance(other, collections_abc.Hashable):
            return self._state == other
        if not isinstance(other, BaseAuxiliaryData):
            return False
        other_compared = type(other)._relationship.compared
        if not other_compared:
            return False
        if isinstance(self, BaseInteractiveAuxiliaryData) != isinstance(
            other, BaseInteractiveAuxiliaryData
        ):
            return False
        self_auxiliary_type = type(self)._base_auxiliary_type
        other_auxiliary_type = type(other)._base_auxiliary_type
        if self_auxiliary_type is other_auxiliary_type:
            if type(self)._relationship == type(other)._relationship:
                return self._state == other._state
        return False

    @final
    def find_with_attributes(self, **attributes):
        # type: (Any) -> Any
        """
        Find first value that matches unique attribute values.

        :param attributes: Attributes to match.

        :return: Value.

        :raises ValueError: No attributes provided or no match found.
        """
        return super(BaseAuxiliaryData, self).find_with_attributes(**attributes)


# noinspection PyAbstractClass
class BaseInteractiveAuxiliaryData(
    BaseAuxiliaryData[T],
    BaseInteractiveData[T],
    BaseInteractiveAuxiliaryStructure[T],
):
    """
    Base interactive auxiliary data.

    Inherits from:
      - :class:`objetto.bases.BaseAuxiliaryData`
      - :class:`objetto.bases.BaseInteractiveData`
      - :class:`objetto.bases.BaseInteractiveAuxiliaryStructure`

    Inherited by:
      - :class:`objetto.data.InteractiveDictData`
      - :class:`objetto.data.InteractiveListData`
      - :class:`objetto.data.InteractiveSetData`
    """

    __slots__ = ()



# -*- coding: utf-8 -*-
"""List operations helper functions."""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Optional, Tuple, Union

__all__ = ["resolve_index", "resolve_continuous_slice", "pre_move"]


def resolve_index(length, index, clamp=False):
    # type: (int, int, bool) -> int
    """
    Resolve index to a positive number.

    :param length: Length of the list.
    :type length: int

    :param index: Input index.
    :type index: int

    :param clamp: Whether to clamp between zero and the length.
    :type clamp: bool

    :return: Resolved index.
    :rtype: int

    :raises IndexError: Index out of range.
    """
    if index < 0:
        index += length
    if clamp:
        if index < 0:
            index = 0
        elif index > length:
            index = length
    elif index < 0 or index >= length:
        error = "index out of range"
        raise IndexError(error)
    return index


def resolve_continuous_slice(length, slc):
    # type: (int, slice) -> Tuple[int, int]
    """
    Resolve continuous slice according to length.

    :param length: Length of the list.
    :type length: int

    :param slc: Continuous slice.
    :type slc: slice

    :return: Index and stop.
    :rtype: tuple[int, int]

    :raises IndexError: Slice is noncontinuous.
    """
    index, stop, step = slc.indices(length)
    if step != 1 or stop < index:
        error = "slice {} is noncontinuous".format(slc)
        raise IndexError(error)
    return index, stop


def pre_move(length, item, target_index):
    # type: (int, Union[slice, int], int) -> Optional[Tuple[int, int, int, int]]
    """
    Perform checks before moving values internally.

    :param length: Length of the list.
    :type length: int

    :param item: Index/slice.
    :type item: int or slice

    :param target_index: Target index.
    :type target_index: int

    :return: None or (index, stop, target index, post index).
    :rtype: None or tuple[int, int, int, int]
    """

    # Resolve slice/index.
    if isinstance(item, slice):
        index, stop = resolve_continuous_slice(length, item)
        if index == stop:
            return None
    else:
        index = resolve_index(length, item)
        stop = index + 1

    # Calculate target index and post index.
    target_index = resolve_index(length, target_index, clamp=True)
    if index <= target_index <= stop:
        return None
    elif target_index > stop:
        post_index = target_index - (stop - index)
    else:
        post_index = target_index

    return index, stop, target_index, post_index



# -*- coding: utf-8 -*-

from itertools import chain
from math import floor

import pytest

from objetto.utils.lazy_import import decorate_path, get_path, import_path


class MyClass(object):
    class MyNestedClass(object):
        pass


def test_import_path():
    assert import_path("math|floor") is floor
    assert import_path("itertools|chain") is chain
    assert import_path(__name__ + "|MyClass") is MyClass
    assert import_path(__name__ + "|MyClass.MyNestedClass") is MyClass.MyNestedClass


def test_get_path():
    assert get_path(floor) == "math|floor"
    assert get_path(chain) == "itertools|chain"
    assert get_path(MyClass) == __name__ + "|MyClass"
    assert get_path(MyClass.MyNestedClass) == __name__ + "|MyClass.MyNestedClass"


def test_decorate_path():
    assert decorate_path("abstractmethod", "abc") == "abc|abstractmethod"
    assert decorate_path(".|abstractmethod", "abc") == "abc|abstractmethod"
    assert decorate_path("abc|abstractmethod", "") == "abc|abstractmethod"
    assert decorate_path(".abc|Mapping", "collections") == "collections.abc|Mapping"

    with pytest.raises(ValueError):
        decorate_path("abstract method|a b c", "xyz")

    with pytest.raises(ValueError):
        decorate_path("abstract method", "abc")

    with pytest.raises(ValueError):
        decorate_path("abstractmethod", "a b c")

    assert decorate_path("..|Counter", "collections.abc") == "collections|Counter"
    assert decorate_path("..._objects.base|BaseObject", "objetto.changes.base") == (
        "objetto._objects.base|BaseObject"
    )


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
"""Dictionary data structures."""

from typing import TYPE_CHECKING, TypeVar, cast, overload

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

from six import iteritems, with_metaclass

from .._bases import final
from .._states import BaseState, DictState
from .._structures import (
    BaseDictStructure,
    BaseDictStructureMeta,
    BaseInteractiveDictStructure,
    KeyRelationship,
    SerializationError,
)
from .bases import (
    BaseAuxiliaryData,
    BaseAuxiliaryDataMeta,
    BaseInteractiveAuxiliaryData,
)

if TYPE_CHECKING:
    from typing import Any, Dict, Iterable, Mapping, Tuple, Type, Union


__all__ = ["DictData", "InteractiveDictData"]


T = TypeVar("T")  # Any type.
KT = TypeVar("KT")  # Key type.
VT = TypeVar("VT")  # Value type.


class DictDataMeta(BaseAuxiliaryDataMeta, BaseDictStructureMeta):
    """Metaclass for `DictData`."""

    @property
    @final
    def _base_auxiliary_type(cls):
        # type: () -> Type[DictData]
        """Base auxiliary container type."""
        return DictData


# noinspection PyTypeChecker
_DD = TypeVar("_DD", bound="DictData")


class DictData(
    with_metaclass(
        DictDataMeta,
        BaseDictStructure[KT, VT],
        BaseAuxiliaryData[KT],
    )
):
    """
    Dictionary data.

    Inherits from:
      - :class:`objetto.bases.BaseDictStructure`
      - :class:`objetto.bases.BaseAuxiliaryData`

    Inherited by:
      - :class:`objetto.data.InteractiveDictData`

    :param initial: Initial values.
    :type initial: collections.abc.Mapping or collections.abc.Iterable[\
tuple[collections.abc.Hashable, Any]]
    """

    __slots__ = ()

    _key_relationship = KeyRelationship()
    """
    Relationship for dictionary keys.

    :type: objetto.data.KeyRelationship
    """

    @classmethod
    @final
    def __make__(cls, state=DictState()):
        # type: (Type[_DD], BaseState) -> _DD
        """
        Make a new dictionary data.

        :param state: Internal state.
        :return: New dictionary data.
        """
        return super(DictData, cls).__make__(state)

    @final
    def __init__(self, initial=()):
        # type: (Union[Mapping[KT, VT], Iterable[Tuple[KT, VT]]]) -> None
        self._init_state(self.__get_initial_state(dict(initial)))

    @classmethod
    @final
    def __get_initial_state(cls, input_values, factory=True):
        # type: (Mapping[KT, VT], bool) -> DictState[KT, VT]
        """
        Get initial state.

        :param input_values: Input values.
        :param factory: Whether to run values through factory.
        :return: Initial state.
        """
        if not cls._key_relationship.passthrough or not cls._relationship.passthrough:
            state = DictState(
                (
                    cls._key_relationship.fabricate_key(k, factory=factory),
                    cls._relationship.fabricate_value(v, factory=factory),
                )
                for k, v in iteritems(input_values)
            )
        else:
            state = DictState(input_values)
        return state

    @final
    def _clear(self):
        # type: (_DD) -> _DD
        """
        Clear all keys and values.

        :return: Transformed.
        :rtype: objetto.data.DictData
        """
        return type(self).__make__()

    @final
    def _set(self, key, value):
        # type: (_DD, KT, VT) -> _DD
        """
        Set value for key.

        :param key: Key.
        :type key: collections.abc.Hashable

        :param value: Value.

        :return: Transformed.
        :rtype: objetto.data.DictData
        """
        cls = type(self)
        key = cls._key_relationship.fabricate_key(key)
        value = cls._relationship.fabricate_value(value)
        return cls.__make__(self._state.set(key, value))

    @final
    def _discard(self, key):
        # type: (_DD, KT) -> _DD
        """
        Discard key if it exists.

        :param key: Key.
        :type key: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.data.DictData
        """
        return type(self).__make__(self._state.discard(key))

    @final
    def _remove(self, key):
        # type: (_DD, KT) -> _DD
        """
        Delete existing key.

        :param key: Key.
        :type key: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.data.DictData
        """
        return type(self).__make__(self._state.remove(key))

    @overload
    def _update(self, __m, **kwargs):
        # type: (_DD, Mapping[KT, VT], VT) -> _DD
        pass

    @overload
    def _update(self, __m, **kwargs):
        # type: (_DD, Iterable[Tuple[KT, VT]], VT) -> _DD
        pass

    @overload
    def _update(self, **kwargs):
        # type: (_DD, VT) -> _DD
        pass

    @final
    def _update(self, *args, **kwargs):
        """
        Update keys and values.
        Same parameters as :meth:`dict.update`.

        :return: Transformed.
        :rtype: objetto.data.DictData
        """
        update = dict(*args, **kwargs)
        cls = type(self)
        if not cls._key_relationship.passthrough or not cls._relationship.passthrough:
            fabricated_update = (
                (
                    cls._key_relationship.fabricate_key(k),
                    cls._relationship.fabricate_value(v),
                )
                for k, v in iteritems(update)
            )
            return cls.__make__(self._state.update(fabricated_update))
        else:
            return cls.__make__(self._state.update(update))

    @classmethod
    @final
    def deserialize(cls, serialized, **kwargs):
        # type: (Type[_DD], Dict, Any) -> _DD
        """
        Deserialize.

        :param serialized: Serialized.
        :type serialized: dict

        :param kwargs: Keyword arguments to be passed to the deserializers.

        :return: Deserialized.
        :rtype: objetto.data.DictData

        :raises objetto.exceptions.SerializationError: Can't deserialize.
        """
        if not cls._relationship.serialized:
            error = "'{}' is not deserializable".format(cls.__name__)
            raise SerializationError(error)
        state = DictState(
            (
                cls._key_relationship.fabricate_key(k, factory=False),
                cls.deserialize_value(v, location=None, **kwargs),
            )
            for k, v in iteritems(serialized)
        )
        return cls.__make__(state)

    @final
    def serialize(self, **kwargs):
        # type: (Any) -> Dict
        """
        Serialize.

        :param kwargs: Keyword arguments to be passed to the serializers.

        :return: Serialized.
        :rtype: dict

        :raises objetto.exceptions.SerializationError: Can't serialize.
        """
        if not type(self)._relationship.serialized:
            error = "'{}' is not serializable".format(type(self).__fullname__)
            raise SerializationError(error)
        return dict(
            (k, self.serialize_value(v, location=None, **kwargs))
            for k, v in iteritems(self._state)
        )

    @property
    @final
    def _state(self):
        # type: () -> DictState[KT, VT]
        """
        Internal state.

        :rtype: objetto.states.DictState
        """
        return cast("DictState", super(BaseDictStructure, self)._state)


class InteractiveDictData(
    DictData[KT, VT],
    BaseInteractiveDictStructure[KT, VT],
    BaseInteractiveAuxiliaryData[KT],
):
    """
    Interactive dictionary data.

    Inherits from:
      - :class:`objetto.data.DictData`
      - :class:`objetto.bases.BaseInteractiveDictStructure`
      - :class:`objetto.bases.BaseInteractiveAuxiliaryData`
    """

    __slots__ = ()



# -*- coding: utf-8 -*-
"""Mutable structures coordinated by an application."""

from .bases import (
    DATA_METHOD_TAG,
    DELETED,
    UNIQUE_ATTRIBUTES_METADATA_KEY,
    BaseAuxiliaryObject,
    BaseAuxiliaryObjectMeta,
    BaseMutableAuxiliaryObject,
    BaseMutableObject,
    BaseObject,
    BaseProxyObject,
    BaseReaction,
    HistoryDescriptor,
    Relationship,
)
from .dict import DictObject, MutableDictObject, ProxyDictObject
from .list import ListObject, MutableListObject, ProxyListObject
from .object import Attribute, Object
from .set import MutableSetObject, ProxySetObject, SetObject

__all__ = [
    "DELETED",
    "UNIQUE_ATTRIBUTES_METADATA_KEY",
    "DATA_METHOD_TAG",
    "Relationship",
    "BaseReaction",
    "HistoryDescriptor",
    "BaseObject",
    "BaseMutableObject",
    "BaseAuxiliaryObjectMeta",
    "BaseAuxiliaryObject",
    "BaseMutableAuxiliaryObject",
    "BaseProxyObject",
    "Attribute",
    "Object",
    "DictObject",
    "MutableDictObject",
    "ProxyDictObject",
    "ListObject",
    "MutableListObject",
    "ProxyListObject",
    "SetObject",
    "MutableSetObject",
    "ProxySetObject",
]



# -*- coding: utf-8 -*-
"""Dummy context manager."""

__all__ = ["DummyContext"]


class DummyContext(object):
    """
    Dummy context manager.

    .. code:: python

        >>> from objetto.utils.dummy_context import DummyContext

        >>> with DummyContext():
        ...     pass
    """

    __slots__ = ()

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False



# -*- coding: utf-8 -*-
"""Base list classes."""

from abc import abstractmethod
from typing import TYPE_CHECKING, TypeVar, overload

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

from qualname import qualname  # type: ignore
from slotted import SlottedMutableSequence, SlottedSequence

from .bases import (
    BaseCollection,
    BaseInteractiveCollection,
    BaseMutableCollection,
    BaseProtectedCollection,
    final,
)

if TYPE_CHECKING:
    from typing import (
        Any,
        Iterable,
        Iterator,
        MutableSequence,
        Optional,
        Sequence,
        Tuple,
        Union,
    )

__all__ = [
    "BaseList",
    "BaseProtectedList",
    "BaseInteractiveList",
    "BaseMutableList",
]


T = TypeVar("T")  # Any type.
T_co = TypeVar("T_co", covariant=True)  # Any type covariant containers.


class BaseList(BaseCollection[T_co], SlottedSequence):
    """
    Base list collection.

    Inherits from:
      - :class:`objetto.bases.BaseCollection`
      - :class:`slotted.SlottedSequence`
      - :class:`typing.Generic`

    Inherited By:
      - :class:`objetto.bases.BaseProtectedList`
    """

    __slots__ = ()

    def __hash__(self):
        """
        Prevent hashing (not hashable by default).

        :raises TypeError: Not hashable.
        """
        error = "unhashable type: '{}'".format(type(self).__fullname__)
        raise TypeError(error)

    @abstractmethod
    def __eq__(self, other):
        # type: (object) -> bool
        """
        Compare for equality.

        :param other: Another object.

        :return: True if equal.
        :rtype: bool

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def __reversed__(self):
        # type: () -> Iterator[T_co]
        """
        Iterate over reversed values.

        :return: Reversed values iterator.
        :rtype: collections.abc.Iterator

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @overload
    @abstractmethod
    def __getitem__(self, index):
        # type: (int) -> T_co
        pass

    @overload
    @abstractmethod
    def __getitem__(self, index):
        # type: (slice) -> Sequence[T_co]
        pass

    @abstractmethod
    def __getitem__(self, index):
        """
        Get value/values at index/from slice.

        :param index: Index/slice.
        :type index: int or slice

        :return: Value/values.
        :rtype: Any or tuple[Any]

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def count(self, value):
        # type: (Any) -> int
        """
        Count number of occurrences of a value.

        :param value: Value.

        :return: Number of occurrences.
        :rtype: int

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def index(self, value, start=None, stop=None):
        # type: (Any, Optional[int], Optional[int]) -> int
        """
        Get index of a value.

        :param value: Value.

        :param start: Start index.
        :type start: int or None

        :param stop: Stop index.
        :type stop: int or None

        :return: Index of value.
        :rtype: int

        :raises ValueError: Provided stop but did not provide start.
        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def resolve_index(self, index, clamp=False):
        # type: (int, bool) -> int
        """
        Resolve index to a positive number.

        :param index: Input index.
        :type index: int

        :param clamp: Whether to clamp between zero and the length.
        :type clamp: bool

        :return: Resolved index.
        :rtype: int

        :raises IndexError: Index out of range.
        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def resolve_continuous_slice(self, slc):
        # type: (slice) -> Tuple[int, int]
        """
        Resolve continuous slice according to length.

        :param slc: Continuous slice.
        :type slc: slice

        :return: Index and stop.
        :rtype: tuple[int, int]

        :raises IndexError: Slice is noncontinuous.
        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()


# noinspection PyCallByClass
type.__setattr__(BaseList, "__hash__", None)  # force non-hashable


# noinspection PyTypeChecker
_BPL = TypeVar("_BPL", bound="BaseProtectedList")


class BaseProtectedList(BaseList[T], BaseProtectedCollection[T]):
    """
    Base protected list collection.

    Inherits from:
      - :class:`objetto.bases.BaseList`
      - :class:`objetto.bases.BaseProtectedCollection`

    Inherited By:
      - :class:`objetto.bases.BaseInteractiveList`
      - :class:`objetto.bases.BaseMutableList`
      - :class:`objetto.bases.BaseListStructure`
    """

    __slots__ = ()

    @abstractmethod
    def _insert(self, index, *values):
        # type: (_BPL, int, T) -> _BPL
        """
        Insert value(s) at index.

        :param index: Index.
        :type index: int

        :param values: Value(s).

        :return: Transformed.
        :rtype: objetto.bases.BaseProtectedList

        :raises ValueError: No values provided.
        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def _append(self, value):
        # type: (_BPL, T) -> _BPL
        """
        Append value at the end.

        :param value: Value.

        :return: Transformed.
        :rtype: objetto.bases.BaseProtectedList
        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def _extend(self, iterable):
        # type: (_BPL, Iterable[T]) -> _BPL
        """
        Extend at the end with iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Transformed.
        :rtype: objetto.bases.BaseProtectedList
        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def _remove(self, value):
        # type: (_BPL, T) -> _BPL
        """
        Remove first occurrence of value.

        :param value: Value.

        :return: Transformed.
        :rtype: objetto.bases.BaseProtectedList

        :raises ValueError: Value is not present.
        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def _reverse(self):
        # type: (_BPL) -> _BPL
        """
        Reverse values.

        :return: Transformed.
        :rtype: objetto.bases.BaseProtectedList

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def _move(self, item, target_index):
        # type: (_BPL, Union[slice, int], int) -> _BPL
        """
        Move values internally.

        :param item: Index/slice.
        :type item: int or slice

        :param target_index: Target index.
        :type target_index: int

        :return: Transformed.
        :rtype: objetto.bases.BaseProtectedList

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def _delete(self, item):
        # type: (_BPL, Union[slice, int]) -> _BPL
        """
        Delete values at index/slice.

        :param item: Index/slice.
        :type item: int or slice

        :return: Transformed.
        :rtype: objetto.bases.BaseProtectedList

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def _update(self, index, *values):
        # type: (_BPL, int, T) -> _BPL
        """
        Update value(s) starting at index.

        :param index: Index.
        :type index: int

        :param values: Value(s).

        :return: Transformed.
        :rtype: objetto.bases.BaseProtectedList

        :raises ValueError: No values provided.
        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()


# noinspection PyTypeChecker
_BIL = TypeVar("_BIL", bound="BaseInteractiveList")


# noinspection PyAbstractClass
class BaseInteractiveList(BaseProtectedList[T], BaseInteractiveCollection[T]):
    """
    Base interactive list collection.

    Inherits from:
      - :class:`objetto.bases.BaseProtectedList`
      - :class:`objetto.bases.BaseInteractiveCollection`

    Inherited By:
      - :class:`objetto.bases.BaseInteractiveListStructure`
      - :class:`objetto.states.ListState`
    """

    __slots__ = ()

    @final
    def insert(self, index, *values):
        # type: (_BIL, int, T) -> _BIL
        """
        Insert value(s) at index.

        :param index: Index.
        :type index: int

        :param values: Value(s).

        :return: Transformed.
        :rtype: objetto.bases.BaseInteractiveList

        :raises ValueError: No values provided.
        """
        return self._insert(index, *values)

    @final
    def append(self, value):
        # type: (_BIL, T) -> _BIL
        """
        Append value at the end.

        :param value: Value.

        :return: Transformed.
        :rtype: objetto.bases.BaseInteractiveList
        """
        return self._append(value)

    @final
    def extend(self, iterable):
        # type: (_BIL, Iterable[T]) -> _BIL
        """
        Extend at the end with iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Transformed.
        :rtype: objetto.bases.BaseInteractiveList
        """
        return self._extend(iterable)

    @final
    def remove(self, value):
        # type: (_BIL, T) -> _BIL
        """
        Remove first occurrence of value.

        :param value: Value.

        :return: Transformed.
        :rtype: objetto.bases.BaseInteractiveList

        :raises ValueError: Value is not present.
        """
        return self._remove(value)

    @final
    def reverse(self):
        # type: (_BIL) -> _BIL
        """
        Reverse values.

        :return: Transformed.
        :rtype: objetto.bases.BaseInteractiveList
        """
        return self._reverse()

    @final
    def move(self, item, target_index):
        # type: (_BPL, Union[slice, int], int) -> _BPL
        """
        Move values internally.

        :param item: Index/slice.
        :type item: int or slice

        :param target_index: Target index.
        :type target_index: int

        :return: Transformed.
        :rtype: objetto.bases.BaseInteractiveList
        """
        return self._move(item, target_index)

    @final
    def delete(self, item):
        # type: (_BPL, Union[slice, int]) -> _BPL
        """
        Delete values at index/slice.

        :param item: Index/slice.
        :type item: int or slice

        :return: Transformed.
        :rtype: objetto.bases.BaseInteractiveList
        """
        return self._delete(item)

    @final
    def update(self, index, *values):
        # type: (_BIL, int, T) -> _BIL
        """
        Update value(s) starting at index.

        :param index: Index.
        :type index: int

        :param values: Value(s).

        :return: Transformed.
        :rtype: objetto.bases.BaseInteractiveList

        :raises ValueError: No values provided.
        """
        return self._update(index, *values)


class BaseMutableList(
    SlottedMutableSequence, BaseProtectedList[T], BaseMutableCollection[T]
):
    """
    Base mutable list collection.

    Inherits from:
      - :class:`slotted.SlottedMutableSequence`
      - :class:`objetto.bases.BaseProtectedList`
      - :class:`objetto.bases.BaseMutableCollection`

    Inherited By:
      - :class:`objetto.bases.BaseMutableListStructure`
      - :class:`objetto.objects.ProxyListObject`
    """

    __slots__ = ()

    @final
    def __iadd__(self, iterable):
        # type: (Iterable[T]) -> MutableSequence[T]
        """
        In place addition.

        :param iterable: Another iterable.
        :type iterable: collections.abc.Iterable

        :return: Added list.
        :rtype: objetto.bases.BaseMutableList
        """
        self._extend(iterable)
        return self

    @overload
    @abstractmethod
    def __getitem__(self, index):
        # type: (int) -> T
        pass

    @overload
    @abstractmethod
    def __getitem__(self, index):
        # type: (slice) -> MutableSequence[T]
        pass

    @abstractmethod
    def __getitem__(self, index):
        """
        Get value/values at index/from slice.

        :param index: Index/slice.
        :type index: int or slice

        :return: Value/values.

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @overload
    @abstractmethod
    def __setitem__(self, index, value):
        # type: (int, T) -> None
        pass

    @overload
    @abstractmethod
    def __setitem__(self, slc, values):
        # type: (slice, Iterable[T]) -> None
        pass

    @abstractmethod
    def __setitem__(self, item, value):
        # type: (Union[int, slice], Union[T, Iterable[T]]) -> None
        """
        Set value/values at index/slice.

        :param item: Index/slice.
        :type item: int or slice

        :param value: Value/values.

        :raises IndexError: Slice is noncontinuous.
        :raises ValueError: Values length does not fit in slice.
        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @overload
    @abstractmethod
    def __delitem__(self, index):
        # type: (int) -> None
        pass

    @overload
    @abstractmethod
    def __delitem__(self, slc):
        # type: (slice) -> None
        pass

    @abstractmethod
    def __delitem__(self, item):
        # type: (Union[int, slice]) -> None
        """
        Delete value/values at index/slice.

        :param item: Index/slice.
        :type item: int or slice

        :raises IndexError: Slice is noncontinuous.
        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def pop(self, index=-1):
        # type: (int) -> T
        """
        Pop value from index.

        :param index: Index.
        :type index: int

        :return: Value.

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @final
    def clear(self):
        # type: () -> None
        """Clear."""
        self._clear()

    @final
    def insert(self, index, *values):
        # type: (int, T) -> None
        """
        Insert value(s) at index.

        :param index: Index.
        :type index: int

        :param values: Value(s).

        :raises ValueError: No values provided.
        """
        self._insert(index, *values)

    @final
    def append(self, value):
        # type: (T) -> None
        """
        Append value at the end.

        :param value: Value.
        """
        self._append(value)

    @final
    def extend(self, iterable):
        # type: (Iterable[T]) -> None
        """
        Extend at the end with iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable
        """
        self._extend(iterable)

    @final
    def remove(self, value):
        # type: (T) -> None
        """
        Remove first occurrence of value.

        :param value: Value.

        :raises ValueError: Value is not present.
        """
        self._remove(value)

    @final
    def reverse(self):
        # type: () -> None
        """Reverse values."""
        self._reverse()

    @final
    def move(self, item, target_index):
        # type: (Union[slice, int], int) -> None
        """
        Move values internally.

        :param item: Index/slice.
        :type item: int or slice

        :param target_index: Target index.
        :type target_index: int
        """
        self._move(item, target_index)

    @final
    def delete(self, item):
        # type: (_BPL, Union[slice, int]) -> None
        """
        Delete values at index/slice.

        :param item: Index/slice.
        :type item: int or slice
        """
        self._delete(item)

    @final
    def update(self, index, *values):
        # type: (int, T) -> None
        """
        Update value(s) starting at index.

        :param index: Index.
        :type index: int

        :param values: Value(s).

        :raises ValueError: No values provided.
        """
        self._update(index, *values)



# -*- coding: utf-8 -*-
"""Set data structures."""

from typing import TYPE_CHECKING, TypeVar, cast

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

from six import with_metaclass

from .._bases import final
from .._states import BaseState, SetState
from .._structures import (
    BaseInteractiveSetStructure,
    BaseSetStructure,
    BaseSetStructureMeta,
    SerializationError,
)
from .bases import (
    BaseAuxiliaryData,
    BaseAuxiliaryDataMeta,
    BaseInteractiveAuxiliaryData,
)

if TYPE_CHECKING:
    from typing import Any, Iterable, List, Type


__all__ = ["SetData", "InteractiveSetData"]


T = TypeVar("T")  # Any type.


class SetDataMeta(BaseAuxiliaryDataMeta, BaseSetStructureMeta):
    """Metaclass for `SetData`."""

    @property
    @final
    def _base_auxiliary_type(cls):
        # type: () -> Type[SetData]
        """Base auxiliary container type."""
        return SetData


# noinspection PyTypeChecker
_SD = TypeVar("_SD", bound="SetData")


class SetData(
    with_metaclass(
        SetDataMeta,
        BaseSetStructure[T],
        BaseAuxiliaryData[T],
    )
):
    """
    Set data.

    Inherits from:
      - :class:`objetto.bases.BaseSetStructure`
      - :class:`objetto.bases.BaseAuxiliaryData`

    Inherited by:
      - :class:`objetto.data.InteractiveSetData`

    :param initial: Initial values.
    :type initial: collections.abc.Iterable
    """

    __slots__ = ()

    @classmethod
    @final
    def __make__(cls, state=SetState()):
        # type: (Type[_SD], BaseState) -> _SD
        """
        Make a new set data.

        :param state: Internal state.
        :return: New set data.
        """
        return super(SetData, cls).__make__(state)

    @classmethod
    @final
    def _from_iterable(cls, iterable):
        # type: (Iterable) -> SetState
        """
        Make set state from iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Set state.
        :rtype: objetto.states.SetState
        """
        return SetState(iterable)

    @final
    def __init__(self, initial=()):
        # type: (Iterable[T]) -> None
        self._init_state(self.__get_initial_state(initial))

    @classmethod
    @final
    def __get_initial_state(cls, input_values, factory=True):
        # type: (Iterable[T], bool) -> SetState[T]
        """
        Get initial state.

        :param input_values: Input values.
        :param factory: Whether to run values through factory.
        :return: Initial state.
        """
        if not cls._relationship.passthrough:
            state = SetState(
                cls._relationship.fabricate_value(v, factory=factory)
                for v in input_values
            )
        else:
            state = SetState(input_values)
        return state

    @final
    def _clear(self):
        # type: (_SD) -> _SD
        """
        Clear all values.

        :return: Transformed.
        :rtype: objetto.data.SetData
        """
        return type(self).__make__()

    @final
    def _add(self, value):
        # type: (_SD, T) -> _SD
        """
        Add value.

        :param value: Value.
        :type value: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.data.SetData
        """
        cls = type(self)
        fabricated_value = cls._relationship.fabricate_value(value)
        return type(self).__make__(self._state.add(fabricated_value))

    @final
    def _discard(self, *values):
        # type: (_SD, T) -> _SD
        """
        Discard value(s).

        :param values: Value(s).
        :type values: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.data.SetData

        :raises ValueError: No values provided.
        """
        return type(self).__make__(self._state.discard(*values))

    @final
    def _remove(self, *values):
        # type: (_SD, T) -> _SD
        """
        Remove existing value(s).

        :param values: Value(s).
        :type value: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.data.SetData

        :raises ValueError: No values provided.
        :raises KeyError: Value is not present.
        """
        return type(self).__make__(self._state.remove(*values))

    @final
    def _replace(self, value, new_value):
        # type: (_SD, T, T) -> _SD
        """
        Replace existing value with a new one.

        :param value: Existing value.
        :type value: collections.abc.Hashable

        :param new_value: New value.
        :type value: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.data.SetData

        :raises KeyError: Value is not present.
        """
        cls = type(self)
        fabricated_new_value = cls._relationship.fabricate_value(new_value)
        return type(self).__make__(self._state.remove(value).add(fabricated_new_value))

    @final
    def _update(self, iterable):
        # type: (_SD, Iterable[T]) -> _SD
        """
        Update with iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable[collections.abc.Hashable]

        :return: Transformed.
        :rtype: objetto.data.SetData
        """
        cls = type(self)
        if not cls._relationship.passthrough:
            fabricated_iterable = (
                cls._relationship.fabricate_value(v) for v in iterable
            )
            return type(self).__make__(self._state.update(fabricated_iterable))
        else:
            return type(self).__make__(self._state.update(iterable))

    @classmethod
    @final
    def deserialize(cls, serialized, **kwargs):
        # type: (Type[_SD], List, Any) -> _SD
        """
        Deserialize.

        :param serialized: Serialized.
        :type serialized: list

        :param kwargs: Keyword arguments to be passed to the deserializers.

        :return: Deserialized.
        :rtype: objetto.data.SetData

        :raises objetto.exceptions.SerializationError: Can't deserialize.
        """
        if not cls._relationship.serialized:
            error = "'{}' is not deserializable".format(cls.__name__)
            raise SerializationError(error)
        state = SetState(
            cls.deserialize_value(v, location=None, **kwargs) for v in serialized
        )
        return cls.__make__(state)

    @final
    def serialize(self, **kwargs):
        # type: (Any) -> List
        """
        Serialize.

        :param kwargs: Keyword arguments to be passed to the serializers.

        :return: Serialized.
        :rtype: list

        :raises objetto.exceptions.SerializationError: Can't serialize.
        """
        if not type(self)._relationship.serialized:
            error = "'{}' is not serializable".format(type(self).__fullname__)
            raise SerializationError(error)
        return sorted(
            (self.serialize_value(v, location=None, **kwargs) for v in self._state),
            key=lambda v: hash(v),
        )

    @property
    @final
    def _state(self):
        # type: () -> SetState[T]
        """
        Internal state.

        :rtype: objetto.states.SetState
        """
        return cast("SetState", super(BaseSetStructure, self)._state)


class InteractiveSetData(
    SetData[T],
    BaseInteractiveSetStructure[T],
    BaseInteractiveAuxiliaryData[T],
):
    """
    Interactive set data.

    Inherits from:
      - :class:`objetto.data.SetData`
      - :class:`objetto.bases.BaseInteractiveSetStructure`
      - :class:`objetto.bases.BaseInteractiveAuxiliaryData`
    """

    __slots__ = ()



# -*- coding: utf-8 -*-
"""Mutable structures coordinated by an application."""

from abc import abstractmethod
from contextlib import contextmanager
from inspect import getmro
from typing import TYPE_CHECKING, TypeVar, cast, overload
from weakref import WeakKeyDictionary

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

from six import integer_types, iteritems, itervalues, string_types, with_metaclass

from .._applications import Application
from .._bases import Base, BaseHashable, BaseMutableCollection, final
from .._changes import Batch
from .._data import BaseAuxiliaryData, BaseData, DataRelationship
from .._states import BaseState, DictState
from .._structures import (
    BaseAuxiliaryStructure,
    BaseAuxiliaryStructureMeta,
    BaseMutableAuxiliaryStructure,
    BaseMutableStructure,
    BaseRelationship,
    BaseStructure,
    BaseStructureMeta,
    make_auxiliary_cls,
)
from ..utils.custom_repr import custom_mapping_repr
from ..utils.recursive_repr import recursive_repr
from ..utils.reraise_context import ReraiseContext
from ..utils.subject_observer import Subject
from ..utils.type_checking import assert_is_instance, assert_is_subclass, import_types
from ..utils.weak_reference import WeakReference

if TYPE_CHECKING:
    from typing import (
        Any,
        Callable,
        Dict,
        Iterator,
        Mapping,
        MutableMapping,
        Optional,
        Set,
        Tuple,
        Type,
        Union,
    )

    from .._applications import Action, Phase, Store
    from .._history import HistoryObject
    from .._states import SetState
    from ..utils.factoring import LazyFactory
    from ..utils.type_checking import LazyTypes

__all__ = [
    "DELETED",
    "UNIQUE_ATTRIBUTES_METADATA_KEY",
    "DATA_METHOD_TAG",
    "Relationship",
    "BaseReaction",
    "HistoryDescriptor",
    "BaseObjectFunctions",
    "BaseObjectMeta",
    "BaseObject",
    "BaseMutableObject",
    "BaseAuxiliaryObjectFunctions",
    "BaseAuxiliaryObjectMeta",
    "BaseAuxiliaryObject",
    "BaseMutableAuxiliaryObject",
    "BaseProxyObject",
]


T = TypeVar("T")  # Any type.

DELETED = object()
"""
Special marker that represents a deleted value.
Can be used when updating :class:`objetto.objects.Object` attributes or
:class:`objetto.objects.DictObject` values.
"""

UNIQUE_ATTRIBUTES_METADATA_KEY = "unique_attributes"
"""Unique attributes index cache metadata key."""

DATA_METHOD_TAG = "__isdatamethod__"
"""Data method tag."""


@final
class Relationship(BaseRelationship):
    """
    Relationship between an object structure and its values.

    Inherits from:
      - :class:`objetto.bases.BaseRelationship`

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :param checked: Whether to perform runtime type check.
    :type checked: bool

    :param module: Module path for lazy types/factories.
    :type module: str or None

    :param factory: Value factory.
    :type factory: str or collections.abc.Callable or None

    :param serialized: Whether should be serialized.
    :type serialized: bool

    :param serializer: Custom serializer.
    :type serializer: str or collections.abc.Callable or None

    :param deserializer: Custom deserializer.
    :type deserializer: str or collections.abc.Callable or None

    :param represented: Whether should be represented.
    :type represented: bool

    :param child: Whether object values should be adopted as children.
    :type child: bool

    :param history: Whether to propagate the history to the child object value.
    :type history: bool

    :param data: Whether to generate data for the value.
    :type data: bool

    :param data_relationship: Data relationship (will be generated if not provided).
    :type data_relationship: objetto.data.DataRelationship or None

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    :raises ValueError: Provided 'serialized' but 'child' is False.
    :raises ValueError: Provided 'history' but 'child' is False.
    :raises ValueError: Provided 'data' but 'child' is False.
    :raises ValueError: Provided 'data_relationship' but 'child' is False
    :raises ValueError: Provided 'data_relationship' but 'data' is False.
    """

    __slots__ = (
        "__child",
        "__history",
        "__data",
        "__data_relationship",
    )

    def __init__(
        self,
        types=(),  # type: LazyTypes
        subtypes=False,  # type: bool
        checked=None,  # type: Optional[bool]
        module=None,  # type: Optional[str]
        factory=None,  # type: LazyFactory
        serialized=None,  # type: Optional[bool]
        serializer=None,  # type: LazyFactory
        deserializer=None,  # type: LazyFactory
        represented=True,  # type: bool
        child=True,  # type: bool
        history=None,  # type: Optional[bool]
        data=None,  # type: Optional[bool]
        data_relationship=None,  # type: Optional[DataRelationship]
    ):
        # type: (...) -> None

        # 'serialized', 'history', and 'data'
        if not child:
            if serialized is not None:
                error = "provided 'serialized' but 'child' is False"
                raise ValueError(error)
            else:
                serialized = False
            if history is not None:
                error = "provided 'history' but 'child' is False"
                raise ValueError(error)
            else:
                history = False
            if data is not None:
                error = "provided 'data' but 'child' is False"
                raise ValueError(error)
            else:
                data = False
        else:
            if serialized is None:
                serialized = True
            else:
                serialized = bool(serialized)
            if history is None:
                history = True
            else:
                history = bool(history)
            if data is None:
                data = True
            else:
                data = bool(data)

        # 'data_relationship'
        if data_relationship is not None:
            if not child:
                error = "provided 'data_relationship' but 'child' is False"
                raise ValueError(error)
            if not data:
                error = "provided 'data_relationship' but 'data' is False"
                raise ValueError(error)

        super(Relationship, self).__init__(
            types=types,
            subtypes=subtypes,
            checked=checked,
            module=module,
            factory=factory,
            serialized=serialized,
            serializer=serializer,
            deserializer=deserializer,
            represented=represented,
        )

        self.__child = bool(child)
        self.__history = bool(history) and self.__child
        self.__data = bool(data) and self.__child
        self.__data_relationship = data_relationship

    def to_dict(self):
        # type: () -> Dict[str, Any]
        """
        Convert to dictionary.

        :return: Dictionary.
        :rtype: Dict[str, Any]
        """
        dct = super(Relationship, self).to_dict()
        dct.update(
            {
                "child": self.child,
                "history": self.history,
                "data": self.data,
                "data_relationship": self.data_relationship,
            }
        )
        return dct

    @property
    def child(self):
        # type: () -> bool
        """
        Whether object values should be adopted as children.

        :rtype: bool
        """
        return self.__child

    @property
    def history(self):
        # type: () -> bool
        """
        Whether to propagate the history to the child object value.

        :rtype: bool
        """
        return self.__history

    @property
    def data(self):
        # type: () -> bool
        """
        Whether to generate data for the value.

        :rtype: bool
        """
        return self.__data

    @property
    def data_relationship(self):
        # type: () -> Optional[DataRelationship]
        """
        Data relationship.

        :rtype: objetto.data.DataRelationship or None
        """
        if self.__data and self.__data_relationship is None:
            types = set()  # type: Set[Union[Type, str]]
            for lazy, typ in zip(self.types, import_types(self.types)):
                if issubclass(typ, BaseObject):
                    if isinstance(lazy, string_types):
                        types.add(lazy + ".Data")
                    else:
                        types.add(typ.Data)
                else:
                    types.add(typ)
            self.__data_relationship = DataRelationship(
                types=types,
                subtypes=self.subtypes,
                checked=False,
                module=self.module,
                factory=None,
                serialized=self.serialized,
                serializer=None,
                deserializer=None,
                represented=self.represented,
                compared=True,
            )
        return self.__data_relationship


# noinspection PyTypeChecker
_BR = TypeVar("_BR", bound="BaseReaction")


class BaseReaction(Base):
    """
    Base method-like that gets called whenever an action is sent through the object.

    Inherits from:
      - :class:`objetto.bases.Base`

    Inherited by:
      - :class:`objetto.reactions.CustomReaction`
      - :class:`objetto.reactions.UniqueAttributes`
      - :class:`objetto.reactions.LimitChildren`
      - :class:`objetto.reactions.Limit`
    """

    __slots__ = ("__hash", "_priority")

    def __init__(self):
        self.__hash = None  # type: Optional[int]
        self._priority = None  # type: Optional[int]

    @abstractmethod
    def __call__(self, obj, action, phase):
        """
        React to actions.

        :param obj: Object.
        :type obj: objetto.bases.BaseObject

        :param action: Action.
        :type action: objetto.objects.Action

        :param phase: Phase.
        :type phase: :data:`objetto.constants.PRE` or :data:`objetto.constants.POST`
        """
        raise NotImplementedError()

    @overload
    def __get__(self, instance, owner):
        # type: (_BR, None, Type[BaseObject]) -> _BR
        pass

    @overload
    def __get__(self, instance, owner):
        # type: (BaseObject, Type[BaseObject]) -> Callable[[Action, Phase], None]
        pass

    @overload
    def __get__(self, instance, owner):
        # type: (_BR, object, type) -> _BR
        pass

    @final
    def __get__(self, instance, owner):
        """
        Get bound reaction method from valid instance or this descriptor otherwise.

        :param instance: Instance.
        :type instance: objetto.bases.BaseObject or None

        :param owner: Owner class.
        :type owner: type[objetto.bases.BaseObject]

        :return: Bound reaction method or this descriptor.
        :rtype: function or objetto.bases.BaseReaction
        """
        if instance is not None:

            def reaction(action, phase):
                # type: (Action, Phase) -> None
                """
                Bound reaction method.

                :param action: Action.
                :param phase: Phase.
                """
                self(instance, action, phase)

            return reaction
        else:
            return self

    @final
    def __hash__(self):
        # type: () -> int
        """
        Get hash.

        :return: Hash.
        :rtype: int
        """
        if self.__hash is None:
            dct = self.to_dict()
            del dct["priority"]
            self.__hash = hash(frozenset(iteritems(dct)))
        return self.__hash

    @final
    def __eq__(self, other):
        # type: (object) -> bool
        """
        Compare with another object for equality.

        :param other: Another object.

        :return: True if considered equal.
        :rtype: bool
        """
        if self is other:
            return True
        if type(self) is not type(other):
            return False
        assert isinstance(other, BaseReaction)
        dct = self.to_dict()
        del dct["priority"]
        if not dct:
            return False
        other_dct = other.to_dict()
        if not other_dct:
            return False
        del other_dct["priority"]
        return dct == other_dct

    @final
    @recursive_repr
    def __repr__(self):
        # type: () -> str
        """
        Get representation.

        :return: Representation.
        :rtype: str
        """
        return custom_mapping_repr(
            self.to_dict(),
            prefix="{}(".format(type(self).__name__),
            template="{key}={value}",
            suffix=")",
            key_repr=str,
        )

    def to_dict(self):
        # type: () -> Dict[str, Any]
        """
        Convert to dictionary.

        :return: Dictionary.
        :rtype: dict[str, Any]
        """
        return {
            "priority": self.priority,
        }

    @final
    def set_priority(self, priority):
        # type: (_BR, Optional[int]) -> _BR
        """
        Set priority and return a new reaction.

        :param priority: Priority.
        :type priority: int or None

        :return: New reaction.
        :rtype: objetto.bases.BaseReaction
        """

        # 'priority'
        if priority is not None:
            with ReraiseContext(TypeError, "'priority' parameter"):
                assert_is_instance(priority, integer_types)

        state = self.__getstate__()
        new_reaction = type(self).__new__(type(self))
        new_reaction.__setstate__(state)
        new_reaction._priority = priority
        return new_reaction

    @property
    @final
    def priority(self):
        # type: () -> Optional[int]
        """
        Priority.

        :rtype: int or None
        """
        return self._priority


# noinspection PyTypeChecker
_HD = TypeVar("_HD", bound="HistoryDescriptor")


@final
class HistoryDescriptor(BaseHashable):
    """
    Descriptor to be used when declaring an :class:`objetto.objects.Object` class.
    When used, every instance of the object class will hold a history that will keep
    track of its changes (and the changes of its children that define a history
    relationship), allowing for easy undo/redo operations.

    If accessed through an instance, the descriptor will return the history object.

    Inherits from:
      - :class:`objetto.bases.BaseHashable`

    :param size: How many changes to remember.
    :type size: int or None

    :raises TypeError: Invalid parameter type.
    """

    __slots__ = ("__size",)

    def __init__(self, size=None):
        # type: (Optional[int]) -> None
        if size is not None:
            with ReraiseContext(TypeError, "'size' parameter"):
                assert_is_instance(size, integer_types)
            size = int(size)
            if size < 0:
                size = 0
        self.__size = size

    @overload
    def __get__(self, instance, owner):
        # type: (_HD, None, Type[BaseObject]) -> _HD
        pass

    @overload
    def __get__(self, instance, owner):
        # type: (BaseObject, Type[BaseObject]) -> HistoryObject
        pass

    @overload
    def __get__(self, instance, owner):
        # type: (_HD, object, type) -> _HD
        pass

    def __get__(self, instance, owner):
        """
        Get history object when accessing from instance or this descriptor otherwise.

        :param instance: Instance.
        :type instance: objetto.bases.BaseObject or None

        :param owner: Owner class.
        :type owner: type[objetto.bases.BaseObject]

        :return: History object or this descriptor.
        :rtype: objetto.history.HistoryObject or objetto.history.HistoryDescriptor
        """
        if instance is not None:
            cls = type(instance)
            if isinstance(instance, BaseObject):
                if getattr(cls, "_history_descriptor", None) is self:
                    history = instance._history
                    assert history is not None
                    return history
        return self

    def __hash__(self):
        # type: () -> int
        """
        Get hash.

        :return: Hash.
        :rtype: int
        """
        return hash(frozenset(iteritems(self.to_dict())))

    def __eq__(self, other):
        # type: (Any) -> bool
        """
        Compare for equality.

        :param other: Another object.

        :return: True if equal.
        :rtype: bool
        """
        if self is other:
            return True
        if not isinstance(other, collections_abc.Hashable):
            return self.to_dict() == other
        if type(self) is not type(other):
            return False
        assert isinstance(other, HistoryDescriptor)
        return self.to_dict() == other.to_dict()

    @recursive_repr
    def __repr__(self):
        # type: () -> str
        """
        Get representation.

        :return: Representation.
        :rtype: str
        """
        return custom_mapping_repr(
            self.to_dict(),
            prefix="{}(".format(type(self).__name__),
            template="{key}={value}",
            suffix=")",
            key_repr=str,
        )

    def to_dict(self):
        # type: () -> Dict[str, Any]
        """
        Convert to dictionary.

        :return: Dictionary.
        :rtype: dict[str, Any]
        """
        return {"size": self.size}

    @property
    def size(self):
        # type: () -> Optional[int]
        """
        How many changes to remember.

        :rtype: int or None
        """
        return self.__size


class BaseObjectFunctions(Base):
    """Base static functions for `BaseObject`."""

    __slots__ = ()

    @staticmethod
    @abstractmethod
    def replace_child_data(store, child, data_location, new_child_data):
        # type: (Store, BaseObject, Any, BaseData) -> Store
        """
        Replace child data.

        :param store: Object's store.
        :param child: Child getting their data replaced.
        :param data_location: Location of the existing child's data.
        :param new_child_data: New child's data.
        :return: Updated object's store.
        """
        raise NotImplementedError()


class BaseObjectInternals(Base):
    """
    Internals for `BaseObject`.

    :param obj: Object.
    :param app: Application.
    """

    __slots__ = ("__obj_ref", "__is_root", "__app", "__subject")

    def __init__(self, obj, app):
        # type: (BaseObject, Application) -> None
        self.__obj_ref = WeakReference(obj)
        self.__is_root = False
        self.__app = app
        self.__subject = Subject()

    def set_root(self):
        # type: () -> None
        """Set object as root."""
        self.__is_root = True

    @property
    def obj_ref(self):
        # type: () -> WeakReference[BaseObject]
        """Weak reference to object."""
        return self.__obj_ref

    @property
    def is_root(self):
        # type: () -> bool
        """Whether object is root."""
        return self.__is_root

    @property
    def app(self):
        # type: () -> Application
        """Application."""
        return self.__app

    @property
    def subject(self):
        # type: () -> Subject
        """Subject."""
        return self.__subject


class BaseObjectMeta(BaseStructureMeta):
    """
    Metaclass for `BaseObject`.

    :raises TypeError: Class has multiple history descriptors.
    """

    __history_descriptor_name = WeakKeyDictionary(
        {}
    )  # type: MutableMapping[BaseObjectMeta, Optional[str]]
    __history_descriptor = WeakKeyDictionary(
        {}
    )  # type: MutableMapping[BaseObjectMeta, Optional[HistoryDescriptor]]
    __reactions = WeakKeyDictionary(
        {}
    )  # type: MutableMapping[BaseObjectMeta, Tuple[BaseReaction, ...]]
    __data_methods = WeakKeyDictionary(
        {}
    )  # type: MutableMapping[BaseObjectMeta, DictState[str, Callable]]

    def __init__(cls, name, bases, dct):
        # type: (str, Tuple[Type, ...], Dict[str, Any]) -> None
        super(BaseObjectMeta, cls).__init__(name, bases, dct)

        # Find history descriptor, data methods, and reactions.
        history_descriptors = {}  # type: Dict[str, HistoryDescriptor]
        reactions = {}  # type: Dict[str, BaseReaction]
        data_methods = {}  # type: Dict[str, Callable]
        for base in reversed(getmro(cls)):
            for member_name, member in iteritems(base.__dict__):
                history_descriptors.pop(member_name, None)
                reactions.pop(member_name, None)
                data_methods.pop(member_name, None)

                # History descriptor.
                if isinstance(member, HistoryDescriptor):
                    history_descriptors[member_name] = member

                # Reaction.
                elif isinstance(member, BaseReaction):
                    reactions[member_name] = member

                # Data method.
                elif callable(member) and getattr(member, DATA_METHOD_TAG, False):
                    data_methods[member_name] = member

        # Can't have more than one history descriptor.
        if len(history_descriptors) > 1:
            error = "class '{}' has multiple history descriptors at {}".format(
                cls.__name__, ", ".join("'{}'".format(n) for n in history_descriptors)
            )
            raise TypeError(error)

        # Store history descriptor and its name.
        history_descriptor_name = None  # type: Optional[str]
        history_descriptor = None  # type: Optional[HistoryDescriptor]
        if len(history_descriptors) == 1:
            history_descriptor_name, history_descriptor = next(
                iteritems(history_descriptors)
            )

        type(cls).__history_descriptor_name[cls] = history_descriptor_name
        type(cls).__history_descriptor[cls] = history_descriptor

        # Store reactions in a tuple, sort them by priority.
        sorted_reactions = tuple(
            sorted(
                itervalues(reactions),
                key=lambda r: (r.priority is None, r.priority),
            )
        )
        type(cls).__reactions[cls] = sorted_reactions

        # Store data methods.
        type(cls).__data_methods[cls] = DictState(data_methods)

    @property
    @abstractmethod
    def _state_factory(cls):
        # type: () -> Callable[..., BaseState]
        """State factory."""
        raise NotImplementedError()

    @property
    @final
    def _serializable_structure_types(cls):
        # type: () -> Tuple[Type[BaseObject]]
        """Serializable structure types."""
        return (BaseObject,)

    @property
    @final
    def _relationship_type(cls):
        # type: () -> Type[BaseRelationship]
        """Relationship type."""
        return Relationship

    @property
    @final
    def _history_descriptor_name(cls):
        # type: () -> Optional[str]
        """History descriptor name or `None`."""
        return type(cls).__history_descriptor_name[cls]

    @property
    @final
    def _history_descriptor(cls):
        # type: () -> Optional[HistoryDescriptor]
        """History descriptor or `None`."""
        return type(cls).__history_descriptor[cls]

    @property
    @final
    def _reactions(cls):
        # type: () -> Tuple[BaseReaction, ...]
        """Reactions sorted by priority."""
        return type(cls).__reactions[cls]

    @property
    @final
    def _data_methods(cls):
        # type: () -> Mapping[str, Callable]
        """Data method functions."""
        return type(cls).__data_methods[cls]

    @property
    @abstractmethod
    def Data(cls):
        # type: () -> Type[BaseData]
        """Data type."""
        raise NotImplementedError()


# noinspection PyTypeChecker
_BO = TypeVar("_BO", bound="BaseObject")


class BaseObject(with_metaclass(BaseObjectMeta, BaseStructure[T])):
    """
    Base object.

    Inherits from:
      - :class:`objetto.bases.BaseStructure`

    Inherited By:
      - :class:`objetto.bases.BaseMutableObject`
      - :class:`objetto.bases.BaseAuxiliaryObject`

    Features:
      - Is a protected structure.

    :param app: Application.
    :type app: objetto.applications.Application
    """

    __slots__ = ("__weakref__", "__")
    __functions__ = BaseObjectFunctions

    def __init__(self, app):
        # type: (Application) -> None
        with ReraiseContext(TypeError, "'app' parameter"):
            assert_is_instance(app, Application)
        self.__ = BaseObjectInternals(self, app)
        app.__.init_object(self)

    @final
    def __copy__(self):
        # type: (_BO) -> _BO
        """
        Get copy by using serialization.

        :return: Copy.
        :rtype: objetto.bases.BaseObject
        """
        return type(self).deserialize(self.serialize())

    @final
    def _hash(self):
        """
        Get hash based on object id.

        :return: Hash based on object id.
        :rtype: int
        """
        return hash(id(self))

    @final
    def _eq(self, other):
        # type: (Any) -> bool
        """
        Compare with another object for identity.

        :param other: Another object.

        :return: True if the same object.
        :rtype: bool
        """
        return self is other

    @abstractmethod
    def _locate(self, child):
        # type: (BaseObject) -> Any
        """
        Locate child object.

        :param child: Child object.
        :type child: objetto.bases.BaseObject

        :return: Location.
        :rtype: collections.abc.Hashable

        :raises ValueError: Could not locate child.
        """
        raise NotImplementedError()

    @abstractmethod
    def _locate_data(self, child):
        # type: (BaseObject) -> Any
        """
        Locate child object's data.

        :param child: Child object.
        :type child: objetto.bases.BaseObject

        :return: Data location.
        :rtype: collections.abc.Hashable

        :raises ValueError: Could not locate child's data.
        """
        raise NotImplementedError()

    @final
    def _in_same_application(self, value):
        # type: (Any) -> bool
        """
        Get whether a value is an object and belongs to the same application as this.

        :param value: Any value or object.

        :return: True if is an object and belongs to the same application.
        :rtype: bool
        """
        return isinstance(value, BaseObject) and self.app is value.app

    @final
    @contextmanager
    def _batch_context(self, name="Batch", **metadata):
        # type: (str, Any) -> Iterator[Batch]
        """
        Batch context.

        :param name: Batch name.
        :type name: str

        :param metadata: Metadata.

        :return: Batch context manager.
        :rtype: contextlib.AbstractContextManager[\
collections.abc.Iterator[objetto.changes.Batch]]
        """
        change = Batch(name=str(name), obj=self, metadata=metadata)
        with self.app.__.batch_context(self, change):
            yield change

    @classmethod
    @abstractmethod
    def deserialize(cls, serialized, app=None, **kwargs):
        # type: (Type[_BO], Dict[str, Any], Application, Any) -> _BO
        """
        Deserialize.

        :param serialized: Serialized.

        :param app: Application (required).
        :type app: objetto.applications.Application

        :param kwargs: Keyword arguments to be passed to the deserializers.

        :return: Deserialized.
        :rtype: objetto.bases.BaseObject

        :raises ValueError: Missing required 'app' attribute.
        :raises objetto.exceptions.SerializationError: Can't deserialize.
        """
        raise NotImplementedError()

    @property
    def _state(self):
        # type: () -> BaseState
        """
        State.

        :rtype: objetto.states.BaseState
        """
        with self.app.__.read_context(self) as read:
            return read().state

    @property
    @final
    def _parent(self):
        # type: () -> Optional[BaseObject]
        """
        Parent object or `None`.

        :rtype: objetto.bases.BaseObject or None
        """
        with self.app.__.read_context(self) as read:
            return read().parent_ref()

    @property
    @final
    def _is_root(self):
        # type: () -> bool
        """
        Whether object is root.

        :rtype: bool
        """
        return self.__.is_root

    @property
    @final
    def _children(self):
        # type: () -> SetState[BaseObject]
        """
        Children objects.

        :rtype: objetto.states.SetState[objetto.bases.BaseObject]
        """
        with self.app.__.read_context(self) as read:
            return read().children

    @property
    @final
    def _history(self):
        # type: () -> Optional[HistoryObject]
        """
        History.

        :rtype: objetto.history.HistoryObject or None
        """
        with self.app.__.read_context(self) as read:
            store = read()
            if store.history is not None:
                return store.history
            provider = store.history_provider_ref()
            if provider is not None:
                assert isinstance(provider, BaseObject)
                return provider._history
        return None

    @property
    @final
    def app(self):
        # type: () -> Application
        """
        Application.

        :rtype: objetto.applications.Application
        """
        return self.__.app

    @property
    def data(self):
        # type: () -> Optional[BaseData[T]]
        """
        Data.

        :rtype: objetto.bases.BaseData or None
        """
        with self.app.__.read_context(self) as read:
            return read().data


# noinspection PyAbstractClass
class BaseMutableObject(BaseObject[T], BaseMutableStructure[T]):
    """
    Base mutable object.

    Inherits from:
      - :class:`objetto.bases.BaseObject`
      - :class:`objetto.bases.BaseMutableStructure`

    Inherited By:
      - :class:`objetto.objects.Object`
      - :class:`objetto.objects.BaseMutableAuxiliaryObject`

    Features:
      - Is an mutable object structure.
    """

    __slots__ = ()


# noinspection PyAbstractClass
class BaseAuxiliaryObjectFunctions(BaseObjectFunctions):
    """Base static functions for `BaseAuxiliaryObject`."""

    __slots__ = ()

    @staticmethod
    def make_data_cls_dct(auxiliary_cls):
        # type: (Type[BaseAuxiliaryObject]) -> Dict[str, Any]
        """
        Make data class member dictionary.

        :param auxiliary_cls: Base auxiliary object class.
        :return: Data class member dictionary.
        """
        return dict(auxiliary_cls._data_methods)


class BaseAuxiliaryObjectMeta(BaseObjectMeta, BaseAuxiliaryStructureMeta):
    """Metaclass for `BaseAuxiliaryObject`."""

    __data_type = WeakKeyDictionary(
        {}
    )  # type: MutableMapping[BaseAuxiliaryObjectMeta, Type[BaseAuxiliaryData]]

    @property
    @abstractmethod
    def _base_auxiliary_type(cls):
        # type: () -> Type[BaseAuxiliaryObject]
        """Base auxiliary object type."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def _base_auxiliary_data_type(cls):
        # type: () -> Type[BaseAuxiliaryData]
        """Base auxiliary data type."""
        raise NotImplementedError()

    @property
    @final
    def Data(cls):
        # type: () -> Type[BaseAuxiliaryData]
        """Data type."""
        mcs = type(cls)

        # Try to get cached data type.
        cls_ = cast("Type[BaseAuxiliaryObject]", cls)
        try:
            data_type = mcs.__data_type[cls]
        except KeyError:
            user_data_type = None
            user_data_type_owner = None
            for base in reversed(getmro(cls)):
                if "Data" in base.__dict__:
                    user_data_type = base.__dict__["Data"]
                    user_data_type_owner = base

            # User-defined data type.
            if user_data_type is not None:
                assert user_data_type_owner is not None
                with ReraiseContext(
                    TypeError,
                    "custom 'Data' class member defined in '{}'".format(
                        user_data_type_owner.__name__
                    ),
                ):
                    assert_is_subclass(user_data_type, cls._base_auxiliary_data_type)
                mcs.__data_type[cls] = data_type = user_data_type

            # Automatically defined data type.
            else:
                data_relationship = cls_._relationship.data_relationship
                if data_relationship is None:
                    data_type = mcs.__data_type[cls] = cls._base_auxiliary_data_type
                else:
                    data_type = mcs.__data_type[cls] = make_auxiliary_cls(
                        cls._base_auxiliary_data_type,
                        data_relationship,
                        qual_name="{}.{}".format(cls.__fullname__, "Data"),
                        module=cls.__module__,
                        unique_descriptor_name=cls._unique_descriptor_name,
                        dct=cls_.__functions__.make_data_cls_dct(cls_),
                    )

        return data_type


# noinspection PyAbstractClass
class BaseAuxiliaryObject(
    with_metaclass(
        BaseAuxiliaryObjectMeta,
        BaseAuxiliaryStructure[T],
        BaseObject[T],
    )
):
    """
    Base auxiliary object.

    Inherits from:
      - :class:`objetto.bases.BaseAuxiliaryStructure`
      - :class:`objetto.bases.BaseObject`

    Inherited By:
      - :class:`objetto.bases.BaseMutableAuxiliaryObject`
      - :class:`objetto.objects.DictObject`
      - :class:`objetto.objects.ListObject`
      - :class:`objetto.objects.SetObject`
    """

    __slots__ = ()
    __functions__ = BaseAuxiliaryObjectFunctions

    _relationship = Relationship()
    """
    Relationship for all locations.

    :type: objetto.objects.Relationship
    """

    @final
    def find_with_attributes(self, **attributes):
        # type: (Any) -> Any
        """
        Find first value that matches unique attribute values.
        This method will be optimized if the auxiliary objects is utiling the
        :class:`objetto.reactions.UniqueAttributes` reaction, which caches unique
        attributes as indexes.

        :param attributes: Attributes to match.

        :return: Value.

        :raises ValueError: No attributes provided or no match found.
        """
        with self.app.__.read_context(self) as read:
            metadata = read().metadata

            # The 'UniqueAttributes' reaction caches children with unique attributes.
            if UNIQUE_ATTRIBUTES_METADATA_KEY in metadata:
                if not attributes:
                    error = "no attributes provided"
                    raise ValueError(error)
                cache = metadata[UNIQUE_ATTRIBUTES_METADATA_KEY]
                if set(cache).issuperset(attributes):
                    match = None
                    for name, value in iteritems(attributes):
                        if value not in cache[name]:
                            break
                        this_match = cache[name][value]
                        if match is not None and this_match is not match:
                            break
                        match = this_match
                    else:
                        return match

            # Fallback to iterating over the state (slower).
            return self._state.find_with_attributes(**attributes)


# noinspection PyAbstractClass
class BaseMutableAuxiliaryObject(
    BaseAuxiliaryObject[T], BaseMutableObject[T], BaseMutableAuxiliaryStructure[T]
):
    """
    Base mutable auxiliary object.

    Inherits from:
      - :class:`objetto.bases.BaseAuxiliaryObject`
      - :class:`objetto.bases.BaseMutableObject`
      - :class:`objetto.bases.BaseMutableAuxiliaryStructure`

    Inherited By:
      - :class:`objetto.objects.MutableDictObject`
      - :class:`objetto.objects.MutableListObject`
      - :class:`objetto.objects.MutableSetObject`
    """

    __slots__ = ()


# noinspection PyTypeChecker
_BPO = TypeVar("_BPO", bound="BaseProxyObject")


class BaseProxyObject(BaseMutableCollection[T]):
    """
    Base auxiliary proxy object.

    Inherits from:
      - :class:`objetto.bases.BaseMutableCollection`

    Inherited By:
      - :class:`objetto.objects.ProxyDictObject`
      - :class:`objetto.objects.ProxyListObject`
      - :class:`objetto.objects.ProxySetObject`

    :param obj: Auxiliary object.
    :type obj: objetto.bases.BaseAuxiliaryObject
    """

    __slots__ = ("__obj",)

    @final
    def __init__(self, obj):
        # type: (BaseAuxiliaryObject[T]) -> None
        self.__obj = obj

    @final
    @recursive_repr
    def __repr__(self):
        # type: () -> str
        """
        Get representation.

        :return: Representation.
        :rtype: str
        """
        return "{}({})".format(type(self).__fullname__, self._obj)

    @final
    def __hash__(self):
        """
        Get hash based on object id.

        :return: Hash based on object id.
        :rtype: int
        """
        return hash(id(self))

    @final
    def __eq__(self, other):
        # type: (Any) -> bool
        """
        Compare with another object for identity.

        :param other: Another object.

        :return: True if the same object.
        :rtype: bool
        """
        return self is other

    @final
    def __len__(self):
        # type: () -> int
        """
        Get count.

        :return: Count.
        :rtype: int
        """
        return self._obj.__len__()

    @final
    def __iter__(self):
        # type: () -> Iterator[T]
        """
        Iterate over.

        :return: Iterator.
        :rtype: collections.abc.Iterator
        """
        for value in self._obj.__iter__():
            yield value

    @final
    def __contains__(self, value):
        # type: (Any) -> bool
        """
        Get whether value is present.

        :param value: Value.

        :return: True if contains.
        :rtype: bool
        """
        return value in self._obj

    @final
    def _clear(self):
        # type: (_BPO) -> _BPO
        """
        Clear.

        :return: Transformed.
        :rtype: objetto.bases.BaseProxyObject
        """
        self._obj._clear()
        return self

    @final
    def find_with_attributes(self, **attributes):
        # type: (Any) -> Any
        """
        Find first value that matches unique attribute values.

        :param attributes: Attributes to match.

        :return: Value that has matching attributes.

        :raises ValueError: No attributes provided or no match found.
        """
        return self._obj.find_with_attributes(**attributes)

    @property
    def _obj(self):
        # type: () -> BaseAuxiliaryObject
        """
        Auxiliary object.

        :rtype: objetto.bases.BaseAuxiliaryObject
        """
        return self.__obj

    @property
    def _state(self):
        # type: () -> BaseState
        """
        State.

        :rtype: objetto.bases.BaseState
        """
        return self._obj._state

    @property
    @final
    def _parent(self):
        # type: () -> Optional[BaseObject]
        """
        Parent object or `None`.

        :rtype: objetto.bases.BaseObject or None
        """
        return self._obj._parent

    @property
    @final
    def _is_root(self):
        # type: () -> bool
        """
        Whether object is root.

        :rtype: bool
        """
        return self._obj._is_root

    @property
    @final
    def _children(self):
        # type: () -> SetState[BaseObject]
        """
        Children objects.

        :rtype: objetto.states.SetState[objetto.bases.BaseObject]
        """
        return self._obj._children

    @property
    @final
    def _history(self):
        # type: () -> Optional[HistoryObject]
        """
        History or `None`.

        :rtype: objetto.history.HistoryObject or None
        """
        return self._obj._history

    @property
    @final
    def app(self):
        # type: () -> Application
        """
        Application.

        :rtype: objetto.applications.Application
        """
        return self._obj.app

    @property
    def data(self):
        # type: () -> Optional[BaseData]
        """
        Data.

        :rtype: objetto.bases.BaseData or None
        """
        return self._obj.data



# -*- coding: utf-8 -*-

import pytest

from objetto.utils.dummy_context import DummyContext


def test_dummy_context():
    with DummyContext():
        pass


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
"""List structures."""

from abc import abstractmethod
from typing import TYPE_CHECKING, TypeVar

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

from six import with_metaclass

from .._bases import BaseInteractiveList, BaseMutableList, BaseProtectedList, final
from ..utils.custom_repr import custom_iterable_repr
from ..utils.recursive_repr import recursive_repr
from .bases import (
    BaseAuxiliaryStructure,
    BaseAuxiliaryStructureMeta,
    BaseInteractiveAuxiliaryStructure,
    BaseMutableAuxiliaryStructure,
)

if TYPE_CHECKING:
    from typing import Any, Iterator, Optional, Tuple

    from .._states import ListState

__all__ = [
    "BaseListStructureMeta",
    "BaseListStructure",
    "BaseInteractiveListStructure",
    "BaseMutableListStructure",
]


T = TypeVar("T")  # Any type.


# noinspection PyAbstractClass
class BaseListStructureMeta(BaseAuxiliaryStructureMeta):
    """Metaclass for `ListStructure`."""


class BaseListStructure(
    with_metaclass(
        BaseListStructureMeta,
        BaseAuxiliaryStructure[T],
        BaseProtectedList[T],
    )
):
    """
    Base list structure.

    Inherits from:
      - :class:`objetto.bases.BaseAuxiliaryStructure`
      - :class:`objetto.bases.BaseProtectedList`

    Inherited By:
      - :class:`objetto.bases.BaseInteractiveListStructure`
      - :class:`objetto.bases.BaseMutableListStructure`
      - :class:`objetto.data.ListData`
      - :class:`objetto.objects.ListObject`
    """

    __slots__ = ()

    @recursive_repr
    def __repr__(self):
        # type: () -> str
        """
        Get representation.

        :return: Representation.
        :rtype: str
        """
        if type(self)._relationship.represented:
            return custom_iterable_repr(
                self._state,
                prefix="{}([".format(type(self).__fullname__),
                suffix="])",
            )
        else:
            return "<{}>".format(type(self).__fullname__)

    @final
    def __reversed__(self):
        # type: () -> Iterator[T]
        """
        Iterate over reversed values.

        :return: Reversed values iterator.
        :rtype: collections.abc.Iterator
        """
        return reversed(self._state)

    @final
    def __len__(self):
        # type: () -> int
        """
        Get value count.

        :return: Value count.
        :rtype: int
        """
        return len(self._state)

    @final
    def __iter__(self):
        # type: () -> Iterator[T]
        """
        Iterate over values.

        :return: Values iterator.
        :rtype: collections.abc.Iterator
        """
        for value in self._state:
            yield value

    @final
    def __contains__(self, value):
        # type: (Any) -> bool
        """
        Get whether value is present.

        :param value: Value.

        :return: True if contains.
        :rtype: bool
        """
        return value in self._state

    @final
    def count(self, value):
        # type: (Any) -> int
        """
        Count number of occurrences of a value.

        :return: Number of occurrences.
        :rtype: int
        """
        return self._state.count(value)

    @final
    def index(self, value, start=None, stop=None):
        # type: (Any, Optional[int], Optional[int]) -> int
        """
        Get index of a value.

        :param value: Value.

        :param start: Start index.
        :type start: int or None

        :param stop: Stop index.
        :type stop: int or None

        :return: Index of value.
        :rtype: int

        :raises ValueError: Provided stop but did not provide start.
        """
        return self._state.index(value, start=start, stop=stop)

    @final
    def resolve_index(self, index, clamp=False):
        # type: (int, bool) -> int
        """
        Resolve index to a positive number.

        :param index: Input index.
        :type index: int

        :param clamp: Whether to clamp between zero and the length.
        :type clamp: bool

        :return: Resolved index.
        :rtype: int

        :raises IndexError: Index out of range.
        """
        return self._state.resolve_index(index, clamp=clamp)

    @final
    def resolve_continuous_slice(self, slc):
        # type: (slice) -> Tuple[int, int]
        """
        Resolve continuous slice according to length.

        :param slc: Continuous slice.
        :type slc: slice

        :return: Index and stop.
        :rtype: tuple[int, int]

        :raises IndexError: Slice is noncontinuous.
        """
        return self._state.resolve_continuous_slice(slc)

    @property
    @abstractmethod
    def _state(self):
        # type: () -> ListState[T]
        """
        Internal state.

        :rtype: objetto.states.ListState

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class BaseInteractiveListStructure(
    BaseListStructure[T],
    BaseInteractiveAuxiliaryStructure[T],
    BaseInteractiveList[T],
):
    """
    Base interactive list structure.

    Inherits from:
      - :class:`objetto.bases.BaseListStructure`
      - :class:`objetto.bases.BaseInteractiveAuxiliaryStructure`
      - :class:`objetto.bases.BaseInteractiveList`

    Inherited By:
      - :class:`objetto.data.InteractiveListData`
    """

    __slots__ = ()


# noinspection PyAbstractClass
class BaseMutableListStructure(
    BaseMutableList[T],
    BaseListStructure[T],
    BaseMutableAuxiliaryStructure[T],
):
    """
    Base mutable list structure.

    Inherits from:
      - :class:`objetto.bases.BaseMutableList`
      - :class:`objetto.bases.BaseListStructure`
      - :class:`objetto.bases.BaseMutableAuxiliaryStructure`

    Inherited By:
      - :class:`objetto.objects.MutableListObject`
    """

    __slots__ = ()



# -*- coding: utf-8 -*-
"""Dictionary structures."""

from abc import abstractmethod
from typing import TYPE_CHECKING, TypeVar, cast

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

from six import iteritems, iterkeys, itervalues, string_types, with_metaclass

from .._bases import (
    BaseHashable,
    BaseInteractiveDict,
    BaseMutableDict,
    BaseProtectedDict,
    final,
)
from .._states import DictState
from ..utils.custom_repr import custom_mapping_repr
from ..utils.factoring import format_factory, import_factory, run_factory
from ..utils.recursive_repr import recursive_repr
from ..utils.reraise_context import ReraiseContext
from ..utils.type_checking import assert_is_instance, format_types, import_types
from .bases import (
    BaseAuxiliaryStructure,
    BaseAuxiliaryStructureMeta,
    BaseInteractiveAuxiliaryStructure,
    BaseMutableAuxiliaryStructure,
)

if TYPE_CHECKING:
    from typing import Any, Dict, Iterator, Optional, Tuple, Type, Union

    from ..utils.factoring import LazyFactory
    from ..utils.type_checking import LazyTypes

__all__ = [
    "KeyRelationship",
    "BaseDictStructureMeta",
    "BaseDictStructure",
    "BaseInteractiveDictStructure",
    "BaseMutableDictStructure",
]


KT = TypeVar("KT")  # Key type.
VT = TypeVar("VT")  # Value type.


@final
class KeyRelationship(BaseHashable):
    """
    Relationship between a dictionary auxiliary structure and their keys.

    Inherits from:
      - :class:`objetto.bases.BaseHashable`

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :param checked: Whether to perform runtime type check.
    :type checked: bool

    :param module: Module path for lazy types/factories.
    :type module: str or None

    :param factory: Key factory.
    :type factory: str or collections.abc.Callable or None

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    """

    __slots__ = (
        "__hash",
        "__types",
        "__subtypes",
        "__checked",
        "__module",
        "__factory",
    )

    def __init__(
        self,
        types=(),  # type: LazyTypes
        subtypes=False,  # type: bool
        checked=None,  # type: Optional[bool]
        module=None,  # type: Optional[str]
        factory=None,  # type: LazyFactory
    ):

        # 'module'
        with ReraiseContext(TypeError, "'module' parameter"):
            assert_is_instance(module, string_types + (None,))
        module = module or None

        # 'types' and 'checked'
        with ReraiseContext((ValueError, TypeError), "'types' parameter"):
            types = format_types(types, module=module)
        if not types:
            if checked:
                error = "did not provide any 'types' but 'checked' is True"
                raise ValueError(error)
            if checked is None:
                checked = False
        else:
            if checked is None:
                checked = True
        checked = bool(checked)

        # 'factory'
        with ReraiseContext((ValueError, TypeError), "'factory' parameter"):
            factory = format_factory(factory, module=module)

        self.__hash = None  # type: Optional[int]
        self.__types = types
        self.__subtypes = bool(subtypes)
        self.__checked = checked
        self.__module = module
        self.__factory = factory

    def __hash__(self):
        # type: () -> int
        """
        Get hash.

        :return: Hash.
        :rtype: int
        """
        if self.__hash is None:
            self.__hash = hash(frozenset(iteritems(self.to_dict())))
        return self.__hash

    def __eq__(self, other):
        # type: (object) -> bool
        """
        Compare with another object for equality.

        :param other: Another object.

        :return: True if considered equal.
        :rtype: bool
        """
        if self is other:
            return True
        if type(self) is not type(other):
            return False
        assert isinstance(other, KeyRelationship)
        return self.to_dict() == other.to_dict()

    @recursive_repr
    def __repr__(self):
        # type: () -> str
        """
        Get representation.

        :return: Representation.
        :rtype: str
        """
        return custom_mapping_repr(
            self.to_dict(),
            prefix="{}(".format(type(self).__name__),
            template="{key}={value}",
            suffix=")",
            sorting=True,
            key_repr=str,
        )

    def to_dict(self):
        # type: () -> Dict[str, Any]
        """
        Convert to dictionary.

        :return: Dictionary.
        :rtype: dict[str, Any]
        """
        return {
            "types": frozenset(import_types(self.types)),
            "subtypes": self.subtypes,
            "checked": self.checked,
            "module": self.module,
            "factory": import_factory(self.factory),
        }

    def fabricate_key(self, key, factory=True, **kwargs):
        # type: (Any, bool, Any) -> Any
        """
        Perform type check and run key through factory.

        :param key: Key.
        :type key: collections.abc.Hashable

        :param factory: Whether to run value through factory.
        :type factory: bool

        :param kwargs: Keyword arguments to be passed to the factory.

        :return: Fabricated value.
        :type key: collections.abc.Hashable
        """
        if factory and self.factory is not None:
            key = run_factory(self.factory, args=(key,), kwargs=kwargs)
        if self.types and self.checked:
            assert_is_instance(key, self.types, subtypes=self.subtypes)
        return key

    @property
    def types(self):
        # type: () -> LazyTypes
        """
        Types.

        :rtype: tuple[str or type]
        """
        return self.__types

    @property
    def subtypes(self):
        # type: () -> bool
        """
        Whether to accept subtypes.

        :rtype: bool
        """
        return self.__subtypes

    @property
    def checked(self):
        # type: () -> bool
        """
        Whether to perform runtime type check.

        :rtype: bool
        """
        return self.__checked

    @property
    def module(self):
        # type: () -> Optional[str]
        """
        Module path for lazy types/factories.

        :rtype: str or None
        """
        return self.__module

    @property
    def factory(self):
        # type: () -> LazyFactory
        """
        Key factory.

        :rtype: str or collections.abc.Callable or None
        """
        return self.__factory

    @property
    def passthrough(self):
        # type: () -> bool
        """
        Whether does not perform type checks and has no factory.

        :rtype: bool
        """
        return (not self.types or not self.checked) and self.factory is None


# noinspection PyAbstractClass
class BaseDictStructureMeta(BaseAuxiliaryStructureMeta):
    """Metaclass for `DictStructure`."""

    def __init__(cls, name, bases, dct):
        super(BaseDictStructureMeta, cls).__init__(name, bases, dct)

        # Check key relationship type.
        cls_ = cast("Type[BaseDictStructure]", cls)
        assert_is_instance(
            cls_._key_relationship,
            cls._key_relationship_type,
            subtypes=False,
        )

    @property
    @final
    def _key_relationship_type(cls):
        # type: () -> Type[KeyRelationship]
        """Relationship type."""
        return KeyRelationship


class BaseDictStructure(
    with_metaclass(
        BaseDictStructureMeta,
        BaseAuxiliaryStructure[KT],
        BaseProtectedDict[KT, VT],
    )
):
    """
    Base dictionary structure.

    Inherits from:
      - :class:`objetto.bases.BaseAuxiliaryStructure`
      - :class:`objetto.bases.BaseProtectedDict`

    Inherited By:
      - :class:`objetto.bases.BaseInteractiveDictStructure`
      - :class:`objetto.bases.BaseMutableDictStructure`
      - :class:`objetto.data.DictData`
      - :class:`objetto.objects.DictObject`
    """

    __slots__ = ()

    _key_relationship = KeyRelationship()
    """
    **Class Attribute**

    Relationship for the keys.

    :type: objetto.objects.KeyRelationship or objetto.data.KeyRelationship
    """

    @recursive_repr
    def __repr__(self):
        # type: () -> str
        """
        Get representation.

        :return: Representation.
        :rtype: str
        """
        if type(self)._relationship.represented:
            return custom_mapping_repr(
                self._state,
                prefix="{}({{".format(type(self).__fullname__),
                suffix="})",
            )
        else:
            return "<{}>".format(type(self).__fullname__)

    @final
    def __reversed__(self):
        # type: () -> Iterator[KT]
        """
        Iterate over reversed keys.

        :return: Reversed keys iterator.
        :rtype: collections.abc.Iterator[collections.abc.Hashable]
        """
        return reversed(self._state)

    @final
    def __getitem__(self, key):
        # type: (KT) -> VT
        """
        Get value for key.

        :param key: Key.
        :type key: collections.abc.Hashable

        :return: Value.

        :raises KeyError: Invalid key.
        """
        return self._state[key]

    @final
    def __len__(self):
        # type: () -> int
        """
        Get key count.

        :return: Key count.
        :rtype: int
        """
        return len(self._state)

    @final
    def __iter__(self):
        # type: () -> Iterator[KT]
        """
        Iterate over keys.

        :return: Key iterator.
        :rtype: collections.abc.Iterator[collections.abc.Hashable]
        """
        for key in self._state:
            yield key

    @final
    def __contains__(self, key):
        # type: (Any) -> bool
        """
        Get whether key is present.

        :param key: Key.
        :type key: collections.abc.Hashable

        :return: True if contains.
        :rtype: bool
        """
        return key in self._state

    @final
    def get(self, key, fallback=None):
        # type: (KT, Any) -> Union[VT, Any]
        """
        Get value for key, return fallback value if key is not present.

        :param key: Key.
        :type key: collections.abc.Hashable

        :param fallback: Fallback value.

        :return: Value or fallback value.
        """
        return self._state.get(key, fallback)

    @final
    def iteritems(self):
        # type: () -> Iterator[Tuple[KT, VT]]
        """
        Iterate over items.

        :return: Items iterator.
        :rtype: collections.abc.Iterator[tuple[collections.abc.Hashable, Any]]
        """
        for key, value in iteritems(self._state):
            yield key, value

    @final
    def iterkeys(self):
        # type: () -> Iterator[KT]
        """
        Iterate over keys.

        :return: Keys iterator.
        :rtype: collections.abc.Iterator[collections.abc.Hashable]
        """
        for key in iterkeys(self._state):
            yield key

    @final
    def itervalues(self):
        # type: () -> Iterator[VT]
        """
        Iterate over values.

        :return: Values iterator.
        :rtype: collections.abc.Iterator
        """
        for value in itervalues(self._state):
            yield value

    @property
    @abstractmethod
    def _state(self):
        # type: () -> DictState[KT, VT]
        """
        Internal state.

        :rtype: objetto.states.DictState

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class BaseInteractiveDictStructure(
    BaseDictStructure[KT, VT],
    BaseInteractiveAuxiliaryStructure[KT],
    BaseInteractiveDict[KT, VT],
):
    """
    Base interactive dictionary structure.

    Inherits from:
      - :class:`objetto.bases.BaseDictStructure`
      - :class:`objetto.bases.BaseInteractiveAuxiliaryStructure`
      - :class:`objetto.bases.BaseInteractiveDict`

    Inherited By:
      - :class:`objetto.data.InteractiveDictData`
    """

    __slots__ = ()


# noinspection PyAbstractClass
class BaseMutableDictStructure(
    BaseMutableDict[KT, VT],
    BaseDictStructure[KT, VT],
    BaseMutableAuxiliaryStructure[KT],
):
    """
    Base mutable dictionary structure.

    Inherits from:
      - :class:`objetto.bases.BaseMutableDict`
      - :class:`objetto.bases.BaseDictStructure`
      - :class:`objetto.bases.BaseMutableAuxiliaryStructure`

    Inherited By:
      - :class:`objetto.objects.MutableDictObject`
    """

    __slots__ = ()
